{$F+}
{$P+,S-,W-,R-}
unit Fedit2;

interface

uses forms, classes, messages, winprocs, wintypes, sysutils, controls, Graphics, dsgnintf,
  typinfo, stdctrls, ToolHelp, clipbrd, holder, insp3, menus, handles, fedutils,
  inifiles, pform;

const
  { grid spacings }
  DefaultXStep=8;
  DefaultYStep=DefaultXStep;
  { Editor constants (user setable) }
  LiveDrag : Boolean=False;
  FullIntersect : Boolean=False;
{  EditMode : Boolean=True;}

{Procedure SetEditMode(Mode : Boolean);}

{ Used when shutting down to ensure that all designers are disposed of etc }
Procedure KillDesigners; far;

type
{  TStandardDelphiWndProc=Function (Handle: hWnd; Msg: word; wParam: word; lParam: Longint ) : Longint;}
  TComponentProc=Procedure(Component : TComponent) of Object;

  TGridRecord=Record
    Show : Boolean;
    XStep : Integer;
    YStep : Integer;
    AutoAlign : Boolean;
    Color : TColor;
  end;

  { Publish a few methods 'after-the-fact' }
  TPublicComponent=class(TComponent)
  public
    Procedure MakeDesigning(Value : Boolean);
    procedure ValidateRename(AComponent: TComponent; const CurName, NewName: string); override;
  end;

  { various states of the form editor }
  TEditState=(esIdle,esWantToInsert,esInsertingNow,esSizing,esSelecting,esWaitForMove,esMoving,
              esDefaultedit);

  TFormEditor2=class(TFormDesigner)
  protected
    FFormloaded : Boolean;    { false until we have finished loading }
    FFormModified : Boolean;
    FFileName : TFileName;
    FFileHandle : Integer;    { A file handle assoicated with the form }
    FAutoHolder : Boolean;    { when true non controls inserted cause a holder to be created auto }
    FShowHandles : Boolean;
    FDestroying : Boolean;
    FSizeHandles : Array[0..7] of TXControlHandle;
    FPopUpMenu : TPopupMenu;
    FSizingHandle : THandleType;
    FTempOwner : TComponent;        { To OWN handles and such }
    FComponentEditor : TComponentEditor;
    FComponentClass : TComponentClass; { to insert a new component }
    FVisibleHandles : Boolean;
    FFreakRules : TIniFile;            { relates to non-form parents of Components }
    FFreakClassRules : TStringList;    { as above }
    FGrid : TGridRecord;               { grid settings }
    FEditState : TEditState;
    FAreaRectangle : TRect;         { Region being selected by mouse }
    FPointInForm : TPoint;          { Form based co-ords of mouse message }
    FPointInScreen : TPoint;        { Screen based coordinates }
    FMoveStart : TPoint;            { Where move started from (for Coords) }
    FMoveDelta : TPoint;            { Increment for last mouse mouse move }
    FMoveOffset : TPoint;           { Difference between current mouse and start move }
    FOldMoveOffset : TPoint;        { Difference on previous go }
    FOldCursorrect : TRect;         { To store and restore cursor slip rect }
{    FApplicationOnMessage : TMessageEvent;}
    FPasteControl : TWinControl;    { Control to paste into }
    FUseControl : TControl;         { A TControl owned by TForm and child of UseWinControl }
    FMessageControl : TControl;
    FSelectedComponents : TList;    { All components selected }
    FSelectedParents : TList;       { FSelectedComponents excluding those whose parents are selected }
    FLastHolder : TComponentHolder; { Last holder make automatically by notification }
    FPastedComponent : TComponent;  { Just been read by ReadComponents }
    FFreakers : TList;              { list of components whose autoparents need checking every designer.modified }
    FNewFreakers : TList;           { A list of components bing pasted to be AutoFreaked together after fixups }
    { Inherited ones }
    Procedure PaintGrid; override;
    Procedure Modified; override;
    Procedure ModifiedForm;
    Procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    Procedure ValidateRename(AComponent: TComponent; const CurName, NewName: string); Override;
    Function GetPrivateDirectory : String; override;
    { Not done seriously }
    function CreateMethod(const Name: string; TypeData: PTypeData): TMethod; override;
    function GetMethodName(const Method: TMethod): string; override;
    procedure GetMethods(TypeData: PTypeData; Proc: TGetStrProc); override;
    function MethodExists(const Name: string): Boolean; override;
    procedure RenameMethod(const CurName, NewName: string); override;
    procedure ShowMethod(const Name: string); override;
    { New ones }
    Function GetValidName(Component : TComponent; OldName : String) : String; virtual;
    procedure ForceValidName(Reader: TReader; Component: TComponent; var Name: string); { This one is an event handler }
    Procedure MakeDesigning(Component : TComponent);
    Procedure MakeNotDesigning(Component : TComponent);
    Procedure FillGridRecordDefaults(var Grid : TGridrecord);
    Procedure PaintFormGrid(Form : TForm; Grid : TGridRecord);
    Procedure EditMouseMove(var Msg: TMessage; var Handled: Boolean);
    Procedure EditLButtonDown(var Msg: TMessage; var Handled: Boolean);
    Procedure EditRButtonDown(var Msg: TMessage; var Handled: Boolean);
    Procedure EditLButtonUp(var Msg: TMessage; var Handled: Boolean);
    Procedure EditRButtonUp(var Msg: TMessage; var Handled: Boolean);
    Procedure EditLButtonDblClk(var Msg: TMessage; var Handled: Boolean);
    Procedure StartInsert(var Msg: TMessage; var Handled: Boolean);
    Procedure MoveInsert;
    Procedure EndInsert;
{    Procedure IdentifyControls(TempWinControl : TWinControl; Pos : TPoint);}
    Function RealizeControl(Control : TControl) : TControl;
    Function RealizePasteControl(Control : TControl) : TWinControl;
    Function ValidatePasteControl(Control : TWinControl) : TWinControl;
    Procedure DrawRectangle(Rectangle : TRect);
    Procedure ClipCursor;
    Procedure UnClipCursor;
    Procedure SetEditState(Value : TEditState);
    Procedure SetEditStateIdle;
    Procedure StartSelecting;
    Procedure UpdateMoveOffsets;
    Procedure MoveSelecting;
    Procedure StopSelecting;
    Procedure PrepareMoving;
    Procedure StartMoving;
    Procedure StopMoving;
    Procedure WaitForMove;
    Procedure DoMoving;
    Procedure DoMoveComponent(Component : TComponent);
    Procedure DoFocusrect(Component : TComponent);
    Procedure DoMovedComponent(Component : TComponent);
    Procedure StartSizing(Component : TComponent);
    Procedure DoSizing(Component : TComponent);
    Procedure StopSizing(Component : TComponent);
    Procedure DoSizeComponent(Component : TComponent);
    Procedure DoSizeRect;
    Function SizeRect(Rectangle : TRect; Offset : TPoint) : TRect;
    Procedure DoSizedComponent(Component : TComponent);
    Procedure ForSelection(Proc : TComponentProc);
    Procedure ForParentSelection(Proc : TComponentProc);
    Procedure AlignPointToGrid(var Point : TPoint);
    Procedure CalculateParents;    { Sticks FSelectedComponents in FSelectedParents except those who have parents selected }
    Property EditState : TEditState Read FEditState Write SetEditState;
    Property VisibleHandles : Boolean Read FVisibleHandles;
    Procedure GoingPlural;
    Procedure GoingSingle;
    Function GetHandleRect(HandleType : THandleType; Component : TComponent) : TRect;
    Function GetMainControl : TControl;
    Function GetMainComponent : TComponent;
    Procedure EditComponents;
    Procedure FreeComponentEditor;
{    Procedure DelayPaintHandles(PaintControl : TControl);}
    Procedure UpdateAll(WinControl : TWinControl);
    Procedure DoDefaultEdit;
    Function MakeHolder(AComponent : TComponent) : TComponentHolder;
    Function FindHolder(AComponent : TComponent) : TControl;
    Procedure ReadComponentsFromStream(Stream : TMemoryStream; Owner, Parent : TComponent);
    Procedure FixPosition(Control : TControl);
    Procedure FixZOrder;
    Procedure PrepareHolders; { set Left and Top to match designinfo }
    Procedure InsertTempControls(MakeHolders : Boolean);
    Procedure RemoveTempControls;
    Procedure DoPopupMenu;
    Procedure AddMenuItem(Caption : String; Tag : Longint; Callback : TNotifyEvent);
    Procedure PerformVerb(Sender : TObject);
    procedure ReadedForPaste(Component: TComponent);
    Procedure SetShowHandles(Value : Boolean);
    Function ShowingHandles : Boolean;
    Procedure formSizePosChanged;
    Procedure SetFilename(Value : TFilename);
    Procedure FreakAutoParents(Component : TComponent; Candidate : TWinControl; Map : Boolean; OwnFirst : Boolean);
    Function tryFreak(Component : TComponent; Visual : TControl; Candidate : TWinControl;
             TryInternal : Boolean; MapCoords : Boolean) : Boolean;
    Procedure CheckFreakers;
    Procedure prepareFreakers;
    Function CanBeFreaked(Component : TComponent) : Boolean;
    Function GetFreakInfo(Component : TComponent) : string;
    Procedure StreamToClipboard(Stream : TMemoryStream);
    Procedure StreamFromClipboard(Stream : TMemoryStream);
    Procedure FreaknewFreakers(PasteControl : TWiNControl);
  public
    Constructor Create;
    Destructor Destroy; override;
    Function IsDesignMsg(Control: TControl; var Message: TMessage): Boolean; override;
    Procedure MenuAlign(Sender : TObject);
    Procedure MenuBringToFront(Sender : TObject);
    Procedure MenuSendToBack(Sender : TObject);
    Procedure MenuGridSpacing(Sender : TObject);
    Procedure LooseDesignerFocus(Commit : Boolean);
    Procedure GetDesignerFocus;
    Property Destroying : Boolean Read FDestroying Write FDestroying;
    Procedure StartEdit(Stream : TStream; EditName : TFilename; FileHandle : Integer); { file can be null }
    Procedure StopEdit;
    Procedure NewSelection(Component : TComponent);
    Procedure AddSelection(Component : TComponent);
    Procedure RemoveSelection(Component : TComponent);
    Procedure ToggleSelection(Component : TComponent);
    Function InSelection(Component : TComponent) : Boolean;
    Procedure ClearSelection;
    Procedure CommitChanges; { for partially edited properties }
    Procedure DrawHandles(Component : TComponent; Visible : Boolean);
    Procedure DrawTags(Component : TComponent; Visible : Boolean; Single : Boolean);
    Procedure FixTags(Component : TComponent; Visible : Boolean);
    Procedure DrawAllHandles(Visible : Boolean);
    Procedure RefreshAllHandles;
    Procedure ShowHandles(Visible : Boolean);
    Procedure PrepareInsert(ComponentClass : TComponentClass);
    Property ComponentList : TList Read FSelectedComponents;
    Property MainControl : TControl Read GetMainControl;
    Property MainComponent : TComponent Read GetMainComponent;
    Function GetEditableComponent(Component : TComponent) : TComponent;
    Procedure Copy;
    Procedure Paste;
    Procedure Delete;
    Procedure Cut;
    Procedure EmergencyQuit;
    Property OptShowHandles : Boolean Read FShowHandles Write SetShowHandles;
    Property FormModified : Boolean Read FFormModified write FFormModified;
    Property FileName : TFileName Read FFileName write SetFileName;
    Property Filehandle : Integer Read FFileHandle Write FFileHandle;
    Property MessageControl : TControl Read FMessageControl Write FMessageControl;
    Property PointInForm : TPoint Read FPointInForm Write FPointInForm;
    Property PointInScreen : TPoint Read FPointInScreen Write FPointInScreen;
    Property SizingHandle : THandleType Read FSizingHandle Write FSizingHandle;
    Procedure SizeMessage;
  end;

  TRegisterProc=Procedure;

implementation

uses Main, gridedit;

{ a list of currently existing designers }
var
  Designers : TList;

{ Makes to be topleft<bottomright}
Procedure FixRectangle(var Rect : TRect);
var
  TR : TRect;
begin
  TR:=Rect;
  If TR.Top>TR.Bottom then begin
    Rect.Bottom:=TR.Top;
    Rect.Top:=TR.Bottom;
  end;
  If TR.Left>TR.Right then begin
    Rect.Left:=TR.Right;
    Rect.Right:=TR.Left;
  end;
end;

{ Lets you reverse dimenstions while sizing and auto select new handle }
{ i.e. size using EAST and size west to get negative dimensions will }
{ result in a noramlized Rect and FHandleType as WEST }
Procedure FixSizeDirection(var Rect : TRect; var FHandleType : THandleType);
var
  TR : TRect;
begin
  TR:=Rect;
  If TR.Top>TR.Bottom then begin
    Rect.Bottom:=TR.Top;
    Rect.Top:=TR.Bottom;
    { Now swap top and bottomness }
    Case FHandleType of
      thtNW: FHandleType:=thtSW;
      thtN: FHandleType:=thtS;
      thtNE: FHandleType:=thtSE;
      thtSW: FHandleType:=thtNW;
      thtS: FHandleType:=thtN;
      thtSE: FHandleType:=thtNE;
    end;
  end;
  If TR.Left>TR.Right then begin
    Rect.Left:=TR.Right;
    Rect.Right:=TR.Left;
    { Now swap left and rightness }
    Case FHandleType of
      thtNW: FHandleType:=thtNE;
      thtW: FHandleType:=thtE;
      thtSW: FHandleType:=thtSE;
      thtNE: FHandleType:=thtNW;
      thtE: FHandleType:=thtW;
      thtSE: FHandleType:=thtSW;
    end;
  end;
end;

Function GetCurrentModuleHandle : THandle;
Var
  taskEntry : TTaskEntry;
  Moduleentry : TModuleEntry;
begin
  TasKEntry.dwsize:=sizeof(taskentry);
  TaskFindHandle(@TaskEntry,GetCurrentTask);
  Result:=TaskEntry.hModule;
end;

Procedure TPublicComponent.MakeDesigning;
begin
  SetDesigning(Value);
end;

Procedure TPublicComponent.ValidateRename;
begin
  Inherited ValidateRename(AComponent,CurName, NewName);
end;

Constructor TFormEditor2.Create;
begin
  Inherited Create;
  Designers.Add(Self);
  FFreakRules:=TIniFile.Create(RTDPMainForm_CSL.FMapName);
  FFreakClassRules:=TStringList.Create;
  FFreakRules.ReadSectionValues('Freakers',FFreakClassRules);
  FFreakers:=TList.Create;
  FNewFreakers:=TList.Create;
  FTempOwner:=TComponent.Create(nil);
  FillGridRecordDefaults(FGrid);
  FSelectedComponents:=TList.Create;
  FSelectedParents:=TList.Create;
  FShowHandles:=True;
end;

Destructor TFormEditor2.Destroy;
var
  c : Integer;
begin
  Destroying:=true;
  { This will stop us updating the screen and such when we empty selection etc }
  Designers.Remove(Self);
  Designers.Pack;
  { Stop editing each form }
  If Form<>nil then
    StopEdit; { Don't care any more, no file prompts, just stop }
  FFreakClassRules.Free;
  FFreakRules.Free;
  FFreakers.Free;
  FNewFreakers.Free;
  FreeComponentEditor;
  FSelectedParents.Free;
  FSelectedComponents.Free;
  FTempOwner.Free;
  Inherited Destroy;
end;

{ This is where we intercept messages BY COOPERATION of the DELPHI VCL }
Function TFormEditor2.IsDesignMsg;
var
  Handled : Boolean;
  PaintControl : TControl;
  WinControl : TWinControl;
  DRect : TRect;
  SR : TRect;
  c : Integer;
begin
{  If GetTickCount-MainForm.GlobalTickCount>4000 then
    EmergencyQuit;}
  { Control is Control receiving message }
  FMessageControl:=Control;
  Handled:=False;

  Case Message.Msg of
    { if its a WMSIZE and belongs to form, we'd better let ourselves know incase they }
    { are in object inspector etc etc }
    WM_SIZE, WM_MOVE: begin
      If (Control=Form) then
        FormSizePosChanged;
    end;
    WM_SYSCOMMAND: begin
      if (Message.WPAram and $FFF0)=SC_Minimize then
        LooseDesignerFocus(true)
      else if (Message.WPAram and $FFF0)=SC_ZOOM then
        getDesignerFocus
      else if (Message.WPAram and $FFF0)=SC_Restore then
        getDesignerFocus;
    end;
    WM_SETFOCUS: begin
      If Control=Form then begin
        GetDesignerFocus;
      end;
    end;
    WM_CLOSE: begin
      { well we can't deal with it here, post a similar message to the main form }
      PostMessage(RTDPMainForm_CSL.Handle,PWM_CloseEdit,0,longint(form));
    end;
    WM_MouseMove, WM_LButtonDown, WM_RButtonDown, WM_LButtonUp, WM_RButtonUp, WM_LBUTTONDBLCLK: begin
      FPointInScreen:=Control.ClientToScreen(TPoint(Message.LParam));
      FPointInForm:=Form.ScreenToClient(FPointInScreen);
      FUseControl:=RealizeControl(Control);
      Case Message.Msg of
        WM_MouseMove: EditMouseMove(Message,Handled);
        WM_LButtonDown: EditLButtonDown(Message,Handled);
        WM_LButtonUp: EditLButtonUp(Message,Handled);
        WM_LButtonDblClk: EditLButtonDblClk(Message,Handled);
        WM_RButtonDown: begin
          If csDesignInteractive in Control.ControlStyle then
            Message.Msg:=WM_LButtonDown
          else
            EditRButtonDown(Message,Handled);
        end;
        WM_RButtonUp: begin
          If csDesignInteractive in Control.ControlStyle then
            Message.Msg:=WM_LButtonUp
          else
            EditRButtonUp(Message,Handled);
        end;
        WM_RButtonDblClk: begin
          If csDesignInteractive in Control.ControlStyle then
            Message.Msg:=WM_LButtonDblClk
          else
            Handled:=True;
        end;
      end;
    end;
  end;
  Result:=Handled;
end;

Procedure TFormEditor2.Paintgrid;
const
  Lock : Boolean=False;
begin
  If Lock then begin
     messagebeep(Word(-1));
     exit;
  end;
  Lock:=True;
  try
    PaintFormGrid(Form,FGrid);
  finally
    Lock:=False;
  end;
end;

Procedure TFormEditor2.Modified;
begin
  { Shame we have to do this every time! Dreadfull waste of time! }
  CheckFreakers;
  Inspector.Renew(Self); { only renew if we are current }
  { Make handles match new positions }
  RefreshAllHandles;
  ModifiedForm;
end;

Procedure TFormEditor2.ModifiedForm;
begin
  if FFormModified<>true then begin
    FFormModified:=True;
    { Somehowcause MAIN FORM to do FixModified }
    RTDPMainForm_CSL.FixEnabled;
  end else
    FFormModified:=True;
end;

Procedure TFormEditor2.Notification;
begin
  { If a component is being inserted, make a holder for it! }
  { If one is being removed... well we should be removing it and therefore the holder }
  If Operation=opInsert then begin
    If (AComponent is TComponent) and not (AComponent is TControl) and FAutoHolder then begin
      FLastHolder:=MakeHolder(AComponent);
      FixZOrder;
      PostMessage(FLastHolder.Handle,PWM_ComponentMoved,0,0);
    end;
  end else If Operation=opRemove then begin
    { if it is a freaking component, remove freaking entry }
    FFreakers.Remove(AComponent);
    { DON'T do FFREAKERS.PACK we can be quite nested and packing during an iteration is BAAADDD }
    FindHolder(AComponent).Free;
    { There may be other fields which require zero'ing as below }
    If FPasteControl=AComponent then FPasteControl:=nil;
  end;
end;

Procedure TFormEditor2.ValidateRename;
begin
end;

Function TFormEditor2.GetPrivateDirectory;
begin
{SAMFIX}
  Result:='';
end;

{ we have to do something here.... not sure what delphi does.... }
function TFormEditor2.CreateMethod;
var
  Method : ^TMethod;
begin
  New(Method);
  Result:=Method^;
end;

function TFormEditor2.GetMethodName;
begin
  Result:='NullName';
end;

procedure TFormEditor2.GetMethods;
begin
  Proc('NoNamesAtAll');
end;

function TFormEditor2.MethodExists;
begin
  Result:=False;
end;

procedure TFormEditor2.RenameMethod;
begin
end;

procedure TFormEditor2.ShowMethod;
begin
{  Writeln('ShowMethod ',name);}
end;

Procedure TFormEditor2.MakeDesigning;
begin
  TPublicComponent(Component).SetDesigning(True);
end;

Procedure TFormEditor2.MakeNotDesigning;
begin
  TPublicComponent(Component).SetDesigning(False);
end;

Procedure TFormEditor2.StartEdit;
var
  Buffer : Array[0..255] of char;
  c : Integer;
  Component : TComponent;
  p,pp : longint;
  Reader : TReader;
begin
  If (Stream<>nil) then begin
    FFilename:=Editname;
    FFilehandle:=Filehandle;
    Form:=TForm.Create(Application);
    Form.Icon.Handle:=LoadIcon(hInstance,'DESIGNICON');
    { We want the form invisible until its loaded... }
    { Bt we can't, we HAVE to make it visble before we enter design }
    { mode, otherwise it can't go visible }
    { So we hide it another way instead!! }
    Form.Width:=0;
    Form.height:=0;
    Form.Visible:=True;
    Form.Designer:=Self;
    MakeDesigning(Form);
    try
      FAutoHolder:=True;
      try
        P:=Stream.Position;
        Stream.ReadResHeader;
      except
        Stream.Position:=P;
      end;
      Stream.ReadComponent(Form);
    except
      FAutoHolder:=False;
      Form.Designer:=nil;
      Form.Free;
      Raise;
    end;
{    Form.Visible:=True;}
    If FFileHandle<0 then Form.Caption:=FFilename;
    Form.HelpContext:=60;
    FFormLoaded:=true;
    FAutoHolder:=False;
    InsertTempControls(False);
    ClearSelection;
    PrepareHolders;
    PrepareFreakers;
    If Designers.Count>0 then begin
      Inspector.Visible:=True;
      Palette.Visible:=True;
    end;
    ShowHandles(True);
    GetDesignerFocus;
  end;
end;

Procedure TFormEditor2.InsertTempControls;
var
  c : Integer;
  Component : TComponent;
begin
  FSizeHandles[0]:=TXControlHandle.Create(FTempOwner,thtN);
  FSizeHandles[1]:=TXControlHandle.Create(FTempOwner,thtNE);
  FSizeHandles[2]:=TXControlHandle.Create(FTempOwner,thtE);
  FSizeHandles[3]:=TXControlHandle.Create(FTempOwner,thtSE);
  FSizeHandles[4]:=TXControlHandle.Create(FTempOwner,thtS);
  FSizeHandles[5]:=TXControlHandle.Create(FTempOwner,thtSW);
  FSizeHandles[6]:=TXControlHandle.Create(FTempOwner,thtW);
  FSizeHandles[7]:=TXControlHandle.Create(FTempOwner,thtNW);
  For c:=0 to 7 do
    FSizeHandles[c].SetParent(Form.Handle);

  { Now create holders for all components }
  If MakeHolders then begin
    c:=0;
    while c<Form.ComponentCount do begin
      Component:=Form.Components[c];
      If not (Component is TControl) then
        { Must be a component, crdate a holder for it }
        MakeHolder(Component);
      inc(c);
    end;
  end;
end;

Procedure TFormEditor2.RemoveTempControls;
var
  c : Integer;
  Control : TControl;
begin
  {Remove control everything that is owned by tempowner, and delete it! }
  while FTempOwner.ComponentCount>0 do begin
    Control:=FTempOwner.Components[0] as TControl;
    Control.Parent:=nil;
    Control.Owner.RemoveComponent(Control);
    Control.Free;
  end;
end;

Procedure TFormEditor2.StopEdit;
begin
  { The property editor might be in the middle of an edit.... }
  { Well, we will abort it, the user should have confirmed the change }
  If (Form<>nil) then begin
    LooseDesignerFocus(false);
    MakeNotDesigning(Form);
    Form.Designer:=nil;
    { form should have been saved before now if they wanted to save it }

    { If csDestroying then Form probably already half gone }
    If not (csDestroying in Form.ComponentState) then
      Form.Release;

    Form:=nil;
    If Designers.Count=1 then begin{ that 1 is me; not quite dead yet }
      Inspector.Visible:=False;
      Palette.Visible:=False;
    end;
  end;
end;

Procedure TFormEditor2.FillgridRecordDefaults;
begin
  Grid.Show:=True;
  Grid.XStep:=DefaultXStep;
  Grid.YStep:=DefaultYStep;
  Grid.Color:=clBlack;
  Grid.AutoAlign:=True;
end;

Procedure TFormEditor2.PaintFormgrid;
var
  X, Y : Integer;
  Area : TRect;
begin
  If FGrid.Show then begin
    Area:=Form.ClientRect;
    { Need some way to do this fast... }
    y:=0;
    while Y<Area.Bottom do begin
      x:=0;
      while x<Area.Right do begin
        { Plot point }
        Form.Canvas.Pixels[x,y]:=Grid.Color;
        inc(x,Grid.XStep);
      end;
      inc(y,Grid.YStep);
    end;
  end;
end;

Procedure TFormEditor2.EditMouseMove;
begin
  case FEditState of
    esSelecting: begin
      MoveSelecting;
      Handled:=True;
    end;
    esInsertingNow: begin
      MoveInsert;
      Handled:=True;
    end;
    esWaitForMove: begin
      WaitForMove;
      Handled:=True;
    end;
    esMoving: begin
      DoMoving;
      Handled:=true;
    end;
    esSizing: begin
      DoSizing(MainComponent);
      Handled:=True;
    end;
  end;
end;

Procedure TFormEditor2.EditLButtonDown;
var
  Pos : TPoint;
  Control : TControl;
  c : Integer;

begin
  FPasteControl:=RealizePasteControl(FMessageControl);

  If Handled=True then begin
  end else begin { Just normal }
    If RTDPMainForm_CSL.Inserting then PrepareInsert(RTDPMainForm_CSL.InsertComponentClass);
    case FEditState of
      esWantToInsert: begin
        StartInsert(Msg,Handled);
      end;
      esIdle: begin
        If (MSG.WParam and (MK_Control or MK_Shift or MK_MButton or MK_RButton))=MK_Shift then begin
          { Are pressing Shift }
          ToggleSelection(FUseControl);
        end else begin
          { Are not pressing shift }
          { If they are clicking on an unselected one, then NewSelection }
          { else ignore, they may just want to start moving }
          If not InSelection(FUseControl) then begin
            NewSelection(FUseControl);
          end;
          If FSelectedComponents.Count>0 then
            EditState:=esWaitForMove;;
        end;
        If FUseControl=Form then { UseControl=Form } { Select Area }
          EditState:=esSelecting;
        Handled:=True;
      end;
    end;
  end;
  Handled:=True; { somehow some messages are slipping though, and why would we not have handled them all? }
end;

Procedure TFormEditor2.EditRButtonDown;
begin
  { Popup the menu }
  Handled:=True;
  DoPopupMenu;
  FPopUpMenu.PopUp(FPointInScreen.x,FPointInScreen.y);
end;

Procedure TFormEditor2.EditLButtonUp;
begin
  EditState:=esIdle;
  Handled:=True;
end;

Procedure TFormEditor2.EditRButtonUp;
begin
  Handled:=True;
end;

Procedure TFormEditor2.EditLButtonDblClk;
begin
  { Default edit on component editor }
  EditState:=esDefaultEdit;
  Handled:=True;
end;

Procedure TFormEditor2.StartInsert;
begin
  ModifiedForm;
  ClearSelection;
  FMoveStart.X:=FPointInForm.X;
  FMoveStart.Y:=FPointInForm.Y;
  If FGrid.AutoAlign then
    AlignpointToGrid(FMoveStart);
  FMoveOffSet.X:=0;
  FMoveOffset.Y:=0;
  ClipCursor;
  EditState:=esInsertingNow;
  FAreaRectangle.Topleft:=TPoint(FMoveStart);
  FAreaRectangle.BottomRight:=TPoint(FMoveStart);
  DrawRectangle(FAreaRectangle);
  Handled:=True;
end;

Procedure TFormEditor2.MoveInsert;
begin
  DrawRectangle(FAreaRectangle);
  FAreaRectangle.BottomRight:=FPointInForm;
  If FGrid.AutoAlign then
    AlignpointToGrid(FAreaRectangle.BottomRight);
  DrawRectangle(FAreaRectangle);
end;

Procedure TFormEditor2.EndInsert;
var
  Control : TControl;
  Component : TComponent;
  P : TPoint;
begin
  try
    DrawRectangle(FAreaRectangle);
    FixRectangle(FAreaRectangle);
    UnClipCursor;
    FPasteControl:=ValidatePasteControl(FPasteControl);
    FAutoHolder:=True;
    try
      Component:=FComponentClass.Create(Form);
    finally
      FAutoHolder:=False;
    end;
{    LockWindowUpdate(Form.Handle);}
    If Component is TControl then begin
      Control:=Component as TControl;
      Control.Parent:=FPasteControl;
      P:=FPasteControl.ScreenToClient(Form.ClientToScreen(FAreaRectangle.TopLeft));
      OffsetRect(FAreaRectangle,-FAreaRectangle.Left,-FAreaRectangle.Top);
      If (abs(FAreaRectangle.Right-FAreaRectangle.Left)>=FGrid.XStep) and
         (abs(FAreaRectangle.Bottom-FAreaRectangle.Top)>=FGrid.YStep) then
        Control.SetBounds(P.X,P.Y,FAreaRectangle.Right,FAreaRectangle.Bottom)
      else
        Control.SetBounds(P.X,P.Y,Control.Width,Control.Height);
    end else begin
      Control:=FindHolder(Component);
      { Map coordinates from form to holders parent }
      { Which is only USUSALLY the form }
      P:=Control.Parent.ScreenToClient(Form.ClientToScreen(FAreaRectangle.TopLeft));
      Control.Setbounds(P.X,P.Y,0,0);
    end;
    { Now if we want we can check for autoparents }
    Component.Name:=GetValidName(Component,'');
    FreakAutoParents(Component,FPasteControl,true,false);
    FixZOrder;
{    LockWindowUpdate(0);}
    NewSelection(Control);
  finally
    FEditState:=esIdle;
    RTDPMainForm_CSL.GotInsert;
  end;
end;

Function TFormEditor2.RealizeControl;
  procedure ShowControlInfo(Tag : String; Control : TControl);
  var
    t : longint;
    p : Tpoint;
  begin
    P:=Control.ScreenToClient(FPointInScreen);
    t:=Control.Perform(CM_DESIGNHITTEST,0,longint(p));
{    Write(format('%3s %6d %20s',[Tag,t,Control.Classname]));
    writeln;}
  end;

var
  TempControl : TControl;
  TempOwner : TComponent;
begin
  Result:=nil;
  If (Control<>Form) and (Control.Owner<>Form) then begin
    TempControl:=Control;
    { Now loop until TempControl's owner is Form }
    repeat
      If TempControl is TComponentHolder then begin
        Result:=TempControl;
        Break;
      end;
      If TempControl.Owner=Form then begin
        Result:=TempControl;
        Break;
      end;
      { Now; we would rather go the owner route... }
{      Tempowner:=TempControl.Owner;
      While (not (TempOwner is TControl)) and (TempOwner<>nil) do
        TempOwner:=TempOwner.Owner;
      If TempOwner<>nil then
        TempControl:=TempOwner as TControl
      else}
       TempControl:=TempControl.Parent;
    until not (TempControl.Parent is TWinControl);
  end else
    Result:=Control;
  if Result=nil then Result:=Form;
  { Result is now the control we can move }
end;

Procedure TFormEditor2.DrawRectangle;
var
  Rect : TRect;
begin
  Form.Canvas.Pen.Mode:=pmNot;
  Form.Canvas.Pen.Width:=3;
  Form.Canvas.Brush.Style:=bsClear;
  FixRectangle(Rect);
  With Rectangle do Form.Canvas.Rectangle(Left,Top,Right,Bottom);
  {Form.Canvas.DrawFocusRect(Rect);}
end;

Procedure TFormeditor2.ClipCursor;
var
  Rect1, Rect2 : TRect;
begin
  GetClipCursor(FOldCursorRect);
  Rect1:=Form.ClientRect;
  Rect2.TopLeft:=Form.ClientToScreen(Rect1.TopLeft);
  Rect2.BottomRight:=Form.ClientToScreen(Rect1.BottomRight);
  WinProcs.ClipCursor(@Rect2);
  SetCapture(Form.Handle);
end;

Procedure TFormEditor2.UnClipCursor;
begin
  ReleaseCapture;
  WinProcs.ClipCursor(@FOldCursorRect);
end;

Procedure TFormEditor2.StartSelecting;
begin
  UpdateAll(Form);
  FAreaRectangle.Topleft:=TPoint(FPointInForm);
  FAreaRectangle.BottomRight:=TPoint(FPointInForm);
  DrawRectangle(FAreaRectangle);
  ClipCursor;
  FEditState:=esSelecting;
end;

Procedure TFormEditor2.MoveSelecting;
var
  c : integer;
  Component : TComponent;
  Control : TControl;
  IR, DR, CR : TRect;
begin
  DrawRectangle(FAreaRectangle);
  FAreaRectangle.BottomRight:=FPointInForm;
  { Now recalculate FSelectedComponents based on truth }
  { Scan through all components, and those in area add, those not delete }
  { Add and delete take care of double adds and double deletes }
  c:=0;
  while c<Form.ComponentCount do begin
    Component:=Form.Components[c];
    { if its a non visual component then nab its holder (which is NOT owned by form) }
    If not (Component is TControl) then begin
      Control:=FindHolder(Component);
      If Control<>nil then Component:=Control;
    end;
    If Component is TControl then begin
      Control:=Component as TControl;
      CR.TopLeft:=Form.ScreenToClient(Control.ClientToScreen(Point(0,0)));
      CR.BottomRight:=Form.ScreenToClient(Control.ClientToScreen(Point(Control.Width,Control.Height)));
      DR:=FAreaRectangle;
      FixRectangle(DR);
      If (IntersectRect(IR,DR,CR)<>0) then begin
        { They intersect! }
        If FullIntersect then begin
          { Little overlap is no good, is it full? }
          If (UnionRect(IR,DR,CR)<>0) and (EqualRect(IR,DR) or EQualRect(IR,CR)) then
            AddSelection(Control)
          else
            RemoveSelection(Control);
        end else
          AddSelection(Control);
      end else begin
        { They don't }
        RemoveSelection(Control);
      end;
    end;
    inc(c);
  end;
  UpdateAll(Form);
  FAreaRectangle.BottomRight:=FPointInForm;
  DrawRectangle(FAreaRectangle);
end;

Procedure TFormEditor2.SetEditState;
begin
  SetEditStateIdle;
  Case Value of
    esIdle: begin
      SetEditStateIdle;
    end;
    esSizing: begin
      StartSizing(MainComponent);
    end;
    esSelecting: begin
      StartSelecting;
    end;
    esWaitForMove: begin
      PrepareMoving;
    end;
    esMoving: begin
      StartMoving;
    end;
  else
    FEditState:=Value;
  end;
end;

Procedure TFormEditor2.SetEditStateIdle;
begin
  Case FEditState of
    esInsertingNow: begin
      EndInsert;
    end;
    esSizing: begin
      StopSizing(MainComponent);
    end;
    esSelecting: begin
      StopSelecting;
    end;
    esDefaultEdit: begin
      DoDefaultEdit;
    end;
    esMoving: begin
      StopMoving;
    end;
  end;
  FEditState:=esIdle;
end;

Procedure TFormEditor2.StopSelecting;
begin
  FEditState:=esidle;
  DrawRectangle(FAreaRectangle);
  UnClipCursor;
end;

Procedure TFormEditor2.PrepareMoving;
begin
  FMoveStart.X:=FPointInForm.X;
  FMoveStart.Y:=FPointInForm.Y;
  FMoveOffSet.X:=0;
  FMoveOffset.Y:=0;
  FEditState:=esWaitForMove;
end;

Procedure TFormEditor2.WaitForMove;
begin
  UpdateMoveOffsets;
  If (abs(FMoveOffset.X)>0) or (abs(FMoveOffset.Y) >0) then begin
    { So we get the full delta effect now! }
    FMoveOffset:=Point(0,0);
    FMoveDelta:=FMoveOffset;
    EditState:=esMoving;
  end;
end;

Procedure TFormEditor2.StartMoving;
begin
  FEditState:=esMoving;
  ShowHandles(False);
  Application.CancelHint;
  If not LiveDrag then
    ForSelection(DoFocusRect);
  ClipCursor;
end;

Procedure TFormEditor2.StopMoving;
begin
  try
    If not LiveDrag then begin
      ForSelection(DoFocusRect);
      ForParentSelection(DoMovedComponent);
    end;
    Modified;
  finally
    FEditState:=esIdle;
    UnclipCursor;
  end;
  ShowHandles(True);
end;

Procedure TFormEditor2.DoMoving;
begin
  If not LiveDrag then
    ForSelection(DoFocusRect);
  UpdateMoveOffsets;
  { Now iterate over controls and offset each one by point }
  If LiveDrag then begin
    ForParentSelection(DoMoveComponent);
{    Modified;}{Surely not...? We will do modified after we finally finish moving... }
  end else
    ForSelection(DoFocusRect);
end;

Procedure TFormEditor2.UpdateMOveOffsets;
begin
  FOldMoveOffset:=FMoveOffset;
  FMoveOffset.X:=FPointInForm.X-FMoveStart.X;
  FMoveOffset.Y:=FPointInForm.Y-FMoveStart.Y;

  { So as to move only by integer grids }
  { This is in line with the Delphi form designer, when moving, don't move TO }
  { grid points, but BY grid points }
  If FGrid.AutoAlign then
    AlignpointToGrid(FMoveOffset);

  FMoveDelta.X:=FMoveOffset.X-FOldMoveOffset.X;
  FMoveDelta.Y:=FMoveOffset.Y-FOldMoveOffset.Y;
end;

Procedure TFormEditor2.DoFocusRect;
var
  rect : TRect;
  point : TPoint;
begin
  Rect.Top:=0;
  Rect.Left:=0;
  Rect.Right:=(Component as TControl).Width;
  Rect.Bottom:=(Component as TControl).Height;
  Point.X:=0;
  Point.Y:=0;
  Point:=Form.ScreenToClient((Component as TControl).ClientToScreen(Point));
  OffsetRect(Rect,point.X,Point.Y);
  OffsetRect(Rect,FMoveOffset.X,FMoveOffset.Y);
  DrawRectangle(Rect);
end;

Procedure TFormEditor2.DoMoveComponent;
begin
  (Component as TControl).SetBounds((Component as TControl).Left+FMoveDelta.X,
                                    (Component as TControl).Top+FMoveDelta.Y,
                                    (Component as TControl).Width,
                                    (Component as TControl).Height);
end;

Procedure TFormEditor2.DoMovedComponent;
begin
  (Component as TControl).SetBounds((Component as TControl).Left+FMoveOffset.X,
                                    (Component as TControl).Top+FMoveOffset.Y,
                                    (Component as TControl).Width,
                                    (Component as TControl).Height);
end;

Procedure TFormEditor2.SizeMEssage;
begin
  If MainControl<>nil then begin
    EditState:=esSizing;
  end;
end;

Procedure TFormEditor2.StartSizing;
begin
  FEditState:=esSizing;
  FMoveStart.X:=FPointInForm.X;
  FMoveStart.Y:=FPointInForm.Y;
  FMoveOffSet.X:=0;
  FMoveOffset.Y:=0;

  ShowHandles(False);
  Application.CancelHint;
  If not LiveDrag then begin
    DoSizeRect;
  end;
  ClipCursor;
end;

Procedure TFormEditor2.StopSizing;
begin
  try
    If not LiveDrag then begin
      DoSizeRect;
      ForParentSelection(DoSizedComponent);
    end;
    Modified;
  finally
    FEditState:=esIdle;
    UnclipCursor;
  end;
  ShowHandles(True);
end;

Procedure TFormEditor2.DoSizing;
begin
  If not LiveDrag then
    DoSizeRect;

  UpdateMoveOffsets;

  If LiveDrag then begin
    DoSizeComponent(FSelectedComponents.Items[0]);
{    Modified;}{We will do modified after we finish sizing }
  end else
    DoSizeRect;
end;

Procedure TFormEditor2.DoSizeComponent;
var
  Control : TControl;
  Rect : TRect;
begin
  Control:=Component as TControl;
  Rect.TopLeft:=Point(Control.Left,Control.Top);
  Rect.Right:=Control.Width+Rect.Left;
  Rect.Bottom:=Control.height+Rect.Top;
  Rect:=SizeRect(Rect,FMoveDelta);
  FixSizeDirection(Rect,FSizingHandle);
  Control.Setbounds(Rect.Left,Rect.Top,Rect.Right-Rect.Left,Rect.Bottom-Rect.Top);
end;

Procedure TFormEditor2.DoSizedComponent;
var
  Control : TControl;
  Rect : TRect;
begin
  Control:=Component as TControl;
  Rect.TopLeft:=Point(Control.Left,Control.Top);
  Rect.Right:=Control.Width+Rect.Left;
  Rect.Bottom:=Control.height+Rect.Top;
  Rect:=SizeRect(Rect,FMoveOffset);
  FixRectangle(Rect);
  Control.Setbounds(Rect.Left,Rect.Top,Rect.Right-Rect.Left,Rect.Bottom-Rect.Top);
end;

Function TFormEditor2.SizeRect;
begin
  With Rectangle do
    Case FSizingHandle of
      thtN:  Result:=Rect(Left,Top+Offset.Y,Right,Bottom);
      thtNE: Result:=Rect(Left,Top+Offset.Y,Right+Offset.X,Bottom);
      thtE:  Result:=Rect(Left,Top,Right+Offset.X,Bottom);
      thtSE: Result:=Rect(Left,Top,Right+Offset.X,Bottom+Offset.Y);
      thtS:  Result:=Rect(Left,Top,Right,Bottom+Offset.Y);
      thtSW: Result:=Rect(Left+Offset.X,Top,Right,Bottom+Offset.Y);
      thtW:  Result:=Rect(Left+Offset.X,Top,Right,Bottom);
      thtNW: Result:=Rect(Left+Offset.X,Top+Offset.Y,Right,Bottom);
    end;
end;

Procedure TFormEditor2.DoSizeRect;
var
  rect : TRect;
  Control : TControl;
begin
  Control:=MainControl;
  Rect.TopLeft:=Form.ScreenToClient(Control.ClientToScreen(Point(0,0)));
  Rect.BottomRight:=Form.ScreenToClient(Control.ClientToScreen(Point(Control.Width,Control.Height)));
  Rect:=SizeRect(Rect,FMoveOffset);
  DrawRectangle(Rect);
end;

Procedure TFormEditor2.ForSelection;
var
  c : integer;
begin
  For c:=0 to FSelectedComponents.Count-1 do
    Proc(FSelectedComponents.Items[c]);
end;

Procedure TFormEditor2.ForParentSelection;
var
  c : integer;
begin
  For c:=0 to FSelectedParents.Count-1 do
    Proc(FSelectedParents.Items[c]);
end;

Procedure TFormEditor2.AlignPointTogrid;
begin
  { Add half the grid spacing, then subtract mod the spacing }
  Point.X:=Point.X+(FGrid.XStep div 2);
  Point.X:=Point.X-(Point.X mod FGrid.XStep);
  Point.Y:=Point.Y+(FGrid.YStep div 2);
  Point.Y:=Point.Y-(Point.Y mod FGrid.YStep);
end;

Procedure TFormEditor2.NewSelection;
begin
  ClearSelection;
  If Component<>Form then
    AddSelection(Component);
end;

Procedure TFormEditor2.AddSelection;
begin
  If Component=Form then Exit;
  If not InSelection(Component) then begin
    If FSelectedComponents.Count=1 then GoingPlural;
    FSelectedComponents.Add(Component);
    If Showinghandles then FixTags(Component,ShowingHandles);
    CalculateParents;
    EditComponents;
  end;
end;

Procedure TFormEditor2.RemoveSelection;
begin
  If InSelection(Component) then
    with FSelectedComponents do begin
      FixTags(Component,False);
      Remove(Component);
      Pack;
      If FSelectedComponents.Count=1 then GoingSingle;
      CalculateParents;
      EditComponents;
    end;
end;

Procedure TFormEditor2.ToggleSelection;
begin
  If InSelection(Component) then
    RemoveSelection(Component)
  else
    AddSelection(Component);
end;

Function TFormEditor2.InSelection;
begin
  Result:=FSelectedComponents.IndexOf(Component)<>-1;
end;

Procedure TFormEditor2.ClearSelection;
begin
  If not Destroying then
    ShowHandles(False);
  FSelectedComponents.Clear;
  If not Destroying then begin
    ShowHandles(True);
    CalculateParents;
    EditComponents;
  end;
end;

Procedure TFormEditor2.DrawHandles;
var
  Control : TControl;
  HW, HH : Integer;

  Procedure DrawHandle(Rect : TRect;HT : THandleType; Visible : Boolean);
  begin
    FSizeHandles[word(HT)].SetState(Form.Handle,Visible,Rect.Left,Rect.Top,Rect.Right-Rect.Left,Rect.Bottom-Rect.Top);
  end;

begin
  Control:=Component as TControl;
  Control.Update;
  HW:=cHandleWidth div 2;
  HH:=cHandleHeight div 2;

  DrawHandle(GetHandleRect(thtNW,Control),thtNW,visible);
  DrawHandle(GetHandleRect(thtN,Control),thtN,visible);
  DrawHandle(GetHandleRect(thtNE,Control),thtNE,visible);

  DrawHandle(GetHandleRect(thtW,Control),thtW,visible);
  DrawHandle(GetHandleRect(thtE,Control),thtE,visible);

  DrawHandle(GetHandleRect(thtSW,Control),thtSW,visible);
  DrawHandle(GetHandleRect(thtS,Control),thtS,visible);
  DrawHandle(GetHandleRect(thtSE,Control),thtSE,visible);

  Control.Update;
end;

Function TFormEditor2.GetHandleRect;
var
  HW, HH : Integer;
  Width, Height : Integer;
  Control : TControl;
begin
  HW:=cHandleWidth div 2;
  HH:=cHandleHeight div 2;
  Control:=Component as TControl;
  Width:=Control.Width;
  Height:=Control.Height;
  Case HandleType of
    thtNW: Result:=Rect(0-HW,0-HH,cHandleWidth,cHandleHeight);
    thtN:  Result:=Rect(((Width-cHandleWidth) Div 2), 0-HH,cHandleWidth,cHandleHeight);
    thtNE: Result:=Rect((Width-cHandleWidth)+HW,0-HH,cHandleWidth,cHandleHeight);
    thtE:  Result:=Rect((Width-cHandleWidth)+HW,(Height-cHandleWidth) Div 2,cHandleWidth,cHandleHeight);
    thtSE: Result:=Rect((Width-cHandleWidth)+HW,(Height-cHandleHeight)+HH,cHandleWidth,cHandleHeight);
    thtS:  Result:=rect((Width-cHandleWidth) Div 2, (Height-cHandleHeight)+HH,cHandleWidth,cHandleHeight);
    thtSW: Result:=Rect(0-HW,(Height-cHandleHeight)+HH,cHandleWidth,cHandleHeight);
    thtW:  Result:=Rect(0-HW,(Height-cHandleHeight) Div 2,cHandleWidth,cHandleHeight);
  end;
  { Patch in the to x,y x,y instead of x,y w,h }
  Result.Bottom:=Result.Top+Result.Bottom;
  Result.Right:=Result.Right+Result.Left;

  Result.TopLeft:=Form.ScreenToClient(Control.ClientToScreen(Result.TopLeft));
  Result.BottomRight:=Form.ScreenToClient(Control.ClientToScreen(Result.BottomRight));
  { Now Clip with ClientRect of form... }
  { otherwise we get autoscroll creep }
  InterSectRect(Result,Result,Form.ClientRect);
end;

Procedure TFormEditor2.GoingSingle;
var
  Component : TComponent;
begin { Convert tags on remaining component to drag tags }
  Component:=(FSelectedComponents.Items[0]);
  DrawTags(Component,False,False);
  DrawTags(Component,True,True);
end;

Procedure TFormEditor2.GoingPlural;
var
  Component : TComponent;
begin { Convert tags on only component to mark tags }
  Component:=(MainComponent);
  DrawTags(Component,False,True);
  DrawTags(Component,True,False);
end;

Procedure TFormEditor2.FixTags;
var
  Single : Boolean;
begin
  Single:=FSelectedComponents.Count=1;
  DrawTags(Component,Visible,Single);
end;

Procedure TFormEditor2.DrawTags;
Var
  DrawControl : TWinControl;
  Control : TControl;
  DrawRect : TRect;
  Width : Integer;
  Height : Integer;
  DC : THandle;

  Procedure DrawTag(X, Y : Integer);
  Var
    Rect : tRect;
  begin
    If Visible then begin
      PatBlt(DC, DrawRect.Left+X,DrawRect.Top+Y,cHandleWidth,cHandleHeight,PATCOPY)
    end else begin
      Rect:=Bounds(DrawRect.Left+X,DrawRect.Top+Y,cHandleWidth,cHandleHeight);
      InvalidateRect(DrawControl.Handle,@Rect,True);
    end;
  end;

Begin
  If Component=Form then exit; { Never draw them for the form }
  If Single then
    DrawHandles(Component,Visible)
  else begin
    If Component is TWinControl then begin
      DrawControl:=(Component as TWinControl);
      DrawRect:=DrawControl.ClientRect;
    end else If Component is TControl then begin
      Control:=(Component as TControl);
      DrawControl:=Control.Parent;
      DrawRect:=Control.ClientRect;
      OffsetRect(DrawRect,Control.Left,Control.Top);
    end else if Component is TComponent then begin
      { should never happen... }
    end;

    { We now need to draw in DrawRect in DrawControl }

    DrawControl.Update;
    Width:=DrawRect.Right-DrawRect.Left;
    Height:=DrawRect.Bottom-DrawRect.Top;

    If Visible Then begin
      DC:=GetDC(DrawControl.Handle);
      SelectObject(DC,GetStockObject(GRAY_BRUSH));
    End;

    DrawTag(0,0);
    DrawTag (Width-cHandleWidth,0);


    DrawTag(0,Height-cHandleHeight);
    DrawTag(Width-cHandleWidth,Height-cHandleHeight);

    if Visible then
      ReleaseDC(DrawControl.Handle, DC);

    DrawControl.Update;
  end;
End;

Procedure TFormEditor2.RefreshAllHandles;
begin
  DrawAllHandles(ShowingHandles);
  UpdateAll(Form);
{  UpdateAll(Form);
  UpdateAll(Form);
  UpdateAll(Form);}
end;

Procedure TFormEditor2.DrawAllHandles;
var
  c : Integer;
begin
  If not visible then
    for c:=0 to 7 do begin
      FSizeHandles[c].SetVisible(Form.Handle,False);
    end;
  For c:=0 to FSelectedComponents.Count-1 do begin
    FixTags(FSelectedComponents.Items[c],Visible);
  end;
end;

Procedure TFormEditor2.ShowHandles;
begin
  FVisibleHandles:=Visible;
  RefreshAllHandles;
end;

Procedure TFormEditor2.PrepareInsert;
begin
  FComponentClass:=ComponentClass;
  EditState:=esWantToInsert;
end;

Procedure TFormEditor2.CalculateParents;
var
  iControl : Integer;
  Component : TComponent;
  Parent, Control : TControl;
  Reject : Boolean;
begin
  FSelectedParents.Clear;
  iControl:=0;
  While iControl<FSelectedComponents.Count do begin
    Component:=FSelectedComponents.Items[iControl];
    If Component is TControl then begin
      Control:=Component as TControl;
      { Is this suitable for scanning parents }
      { If it is a form, ditch it }
      If (Control<>Form) then begin
        { Now scan for all controls parents }
        Parent:=Control.Parent;
        Reject:=False;
        while (Parent<>Form) and (Parent<>nil) do begin
          If InSelection(Parent) then begin
            { Yikes! Found it - forget it }
            Reject:=True;
            break;
          end;
          Parent:=Parent.Parent;
        end;

        If not Reject then begin
          FSelectedParents.Add(Control);
        end;
      end;
    end else begin
      { a component - well well }
      FSelectedParents.Add(Component);
    end;
    inc(iControl);
  end;
end;

Procedure TFormEditor2.EditComponents;
var
  ComponentList : TList;
  c : Integer;
begin
  ComponentList:=TList.Create;
  try
    If FSelectedComponents.Count=0 then
      ComponentList.Add(Form)
    else begin
      c:=0; while c<FSelectedComponents.Count do begin
        ComponentList.Add(GetEditableComponent(FSelectedComponents[c]));
        inc(c);
      end;
    end;
    Inspector.UseComponents(ComponentList,Self);
  except
    ComponentList.Free; { This is usually done by Inspector }
    Raise;
  end;
end;

Procedure TFormEditor2.FreecomponentEditor;
begin
  If FComponentEditor<>nil then begin
    FComponentEditor.Free;
    FComponentEditor:=nil;
  end;
end;

Function TFormeditor2.getMainControl;
var
  Component : TComponent;
begin
  Component:=GetMainComponent;
  If Component is TControl then
    Result:=Component as TControl
  else
    Result:=nil;
end;

Function TFormEditor2.GetMainComponent;
begin
  If FSelectedComponents.Count>0 then
    Result:=FSelectedComponents[0]
  else
    Result:=nil;
end;

Procedure TFormEditor2.AddMenuItem;
Var
  MenuItem : tMenuItem;
begin
  MenuItem:=tMenuItem.Create(FPopupMenu);
  MenuItem.Caption:=Caption;
  MenuItem.Tag:=Tag;
  MenuItem.OnClick:=Callback;
  FPopupMenu.Items.Add(MenuItem);
end;

Procedure TFormEditor2.DoPopupMenu;
var
  C : Integer;
  VerbCount : Integer;
  Verb : String;

begin
  FreeComponentEditor;
  If FPopupMenu<>nil then begin
    FPopupMenu.Free;
    FPopupMenu:=nil;
  end;
  FPopUpMenu:=TPopupMenu.Create(FTempOwner);
  If FSelectedComponents.Count=1 then begin
    FComponentEditor:=GetComponentEditor(FSelectedComponents[0],Self);
    VerbCount:=FComponentEditor.GetVerbCount;
    c:=0; while c<VerbCount do begin
      AddMenuItem(FComponentEditor.GetVerb(c),c,PerformVerb);
      inc(c);
    end;
    If c<>0 then
      AddMenuItem('-',0,nil);
  end;
  { Now we can add our menu Options }
  AddMenuItem('Align To &Grid',0,MenuAlign);
  AddMenuItem('Bring to &Front',0,MenuBringToFront);
  AddMenuItem('Send To &Back',0,MenuSendToBack);
  AddMenuItem('-',0,nil);
  AddMenuItem('Grid &Spacing',0,MenuGridSpacing);
end;

Procedure TFormEditor2.PerformVerb;
begin
  If (FComponentEditor=nil) or (FPopUpMenu=nil) then MessageBeep(Word(-1)) else begin
    FComponentEditor.ExecuteVerb((Sender as TMenuItem).Tag);
  end;
end;

Procedure TFormEditor2.MenuGridSpacing;
begin
  GridEditor.Grid:=FGrid;
  If GridEditor.ShowModal=mrOK then begin
    FGrid:=GridEditor.Grid;
    Form.Repaint;                     
  end;
end;

Procedure TFormEditor2.MenuAlign;
var
  C : integer;
  Component : TComponent;
  Control : TControl;
  P1, P2 : TPoint;
begin
  { Align all points to grid }
  c:=FSelectedParents.Count-1; while c>=0 do begin
    Component:=FSelectedParents[c];
    If (Component is TControl) then begin
      Control:=(Component as TControl);
      P1:=Point(Control.Left,Control.Top);
      P2:=Point(Control.Width,Control.Height);
      AlignPointToGrid(P1);
      AlignPointToGrid(P2);
      { Note we rounded width and height not bottom and right, but its all the same... }
      Control.SetBounds(P1.X,P1.Y,P2.X,P2.Y);
    end;
    dec(c);
  end;
end;

Procedure TFormEditor2.MenuBringToFront;
var
  C : integer;
  Component : TComponent;
begin
  { Bring in opposite selection Order so we preserve Z Order from selection order }
  {SAMFIX parents only or all selected components.... hmmm.... }
  c:=FSelectedParents.Count-1; while c>=0 do begin
    Component:=FSelectedParents[c];
    If (Component is TControl) then (Component as TControl).BringToFront;
    dec(c);
  end;
  FixZOrder;
end;

Procedure TFormEditor2.MenuSendToBack;
var
  C : integer;
  Component : TComponent;
begin
  { Bring in opposite selection Order so we preserve Z Order from selection order }
  {SAMFIX parents only or all selected components.... hmmm.... }
  c:=FSelectedParents.Count-1; while c>=0 do begin
    Component:=FSelectedParents[c];
    If (Component is TControl) then (Component as TControl).SendToBack;
    dec(c);
  end;
  FixZOrder;
end;

Procedure TFormEditor2.FixPosition;
{ also make sure we are not negative coords }
{ If we have same x,y as another sibling control, add grid until we don't }
var
  c : Integer;
  Sibling : TControl;
  X, Y : Integer;
  P : TPoint;
begin
  X:=FGrid.XStep; If X=0 then X:=DefaultXStep;
  Y:=FGrid.YStep; If Y=0 then Y:=DefaultYStep;
  P.X:=Control.Left;
  P.Y:=Control.Top;
  If P.Y<0 then P.Y:=0;
  If P.X<0 then P.X:=0;
  c:=0; while c<Control.Parent.ControlCount do begin
    Sibling:=Control.Parent.Controls[c];
    If (Control<>Sibling) and (P.X=Sibling.Left)
       and (P.Y=Sibling.Top) then begin
         P.X:=P.X+X;
         P.Y:=P.Y+Y;
         { Start again! }
         c:=0;
         continue;
       end;
    inc(c);
  end;
  { P is now not ontop of anything... }
  If P.X+Control.Width>Control.Parent.ClientWidth then
    P.X:=Control.Parent.ClientWidth-Control.Width;
  If P.X<0 then P.X:=0;
  If P.Y+Control.Height>Control.Parent.ClientHeight then
    P.Y:=Control.Parent.ClientHeight-Control.height;
  If P.Y<0 then P.Y:=0;
  { P is now within bounds }
  Control.SetBounds(P.X,P.Y,Control.Width,Control.Height);
  { NOTE: These last changes may have broken earlier conditions, but so what }
  { we did our best.... sort of }
end;

Procedure TFormEditor2.DoDefaultEdit;
var
  ComponentEditor : TComponentEditor;
  Component : TComponent;
begin
  FEditState:=esidle;
  Component:=GetMainComponent;
  If Component<>nil then begin
    ComponentEditor:=GetComponenteditor(Component,Self);
    If Componenteditor<>nil then
      ComponentEditor.Edit;
    ComponentEditor.Free;
  end;
end;

Procedure TFormEditor2.ForceValidName;
begin
  If Component.Owner=Form then { otherwise we have no interest! This test is important as we }
                               { will get called for SubComponents of TArrays etc and mayhap notebooks! }
    Name:=GetValidName(Component,Name);
end;

Procedure TFormEditor2.UpdateAll;
var
  Rect : TRect;
  Handle : THandle;
begin
  if Destroying then exit;
  Handle:=WinControl.Handle;
  GetWindowRect(Handle,Rect);
  winprocs.SCreenToClient(Handle,Rect.TopLeft);
  winprocs.SCreenToClient(Handle,Rect.BottomRight);
  RedrawWindow(WinControl.Handle,@Rect,0,RDW_UpdateNow or RDW_AllChildren);
end;

Function TFormEditor2.RealizePasteControl;
var
  TempControl : TControl;
begin
  TempControl:=Control;
  repeat
    If not (TempControl is TWinControl) then
      TempControl:=RealizeControl(TempControl.Parent) as TWinControl
    else
      TempControl:=(TempControl as TWinControl);
    { Now pop up levels until we find csAcceptControls }
    { This will always happen at FORm level }
    If (csAcceptsControls in TempControl.ControlStyle) then break;
    TempControl:=TempControl.Parent;
  until false;
  Result:=TempControl as TWinControl;
end;

Function TformEditor2.validatePasteControl;
begin
  if (FSelectedComponents.Count<=1) and (control<>nil) then
    Result:=Control
  else
    Result:=Form;
end;

Function TFormEditor2.GetValidName;
var
  Name, BaseName : String;
  Index : Integer;
  Success : Boolean;
begin
  { Return a valid name suitable for assign to the component }
  BaseName:=NiceClassName(Component.Classtype);
  Index:=1;
  repeat
    try
      Name:=Format('%s%d',[BaseName,Index]);
      TPublicComponent(Form).ValidateRename(Component,'',Name);
      Success:=True;
    except
      On E: EComponentError do begin
        inc(index);
        Success:=False;
      end else Raise;
    end;
  until Success;
  Result:=Name;
end;

Function TFormEditor2.GetEditableComponent;
begin
  If Component is TComponentHolder then
    Result:=(Component as TComponentHolder).Component
  else
    Result:=Component;
end;

Function TFormEditor2.MakeHolder;
var
  Control : TComponentHolder;
  NiceHint : String;
  Point : TPoint;
begin
  Control:=TComponentHolder.Create(FTempOwner);
  NiceHint:=NiceClassName(AComponent.Classtype);
  Control.Hint:=NiceHint+'|'+'A token to represent the '+nicehint+' non-visual component';
  Control.showhint:=true;
  (Control as TComponentHolder).Component:=AComponent;
  Point.X:=HiWord(AComponent.DesignInfo);
  Point.Y:=LoWord(AComponent.DesignInfo);

  Control.SetBounds(Point.X,Point.Y,0,0); { last two are overridden }
  Control.Parent:=Form;

  MakeNotDesigning(Control); { Never designing or our hint line won't work }
                             { we hand wndproc to parent form by hand }
  Result:=Control;
end;

Function TFormEditor2.FindHolder;
var
  c : integer;
  Component : TComponent;
  Control : TComponentHolder;
begin
  Result:=nil;
  If AComponent=nil then exit;
  c:=0; while c<FTempOwner.ComponentCount do begin
    Component:=FTempOwner.Components[c];
    If Component is TComponentHolder then begin
      Control:=Component as TComponentHolder;
      If Control.Component=AComponent then begin
        Result:=Control;
        Break;
      end;
    end;
    inc(c);
  end;
end;

Procedure TFormEditor2.Copy;
var
  c, p : Integer;
  BinaryStream, TextStream : TMemoryStream;
  Component : TComponent;
  Control : TControl;
  Holder : TComponentHolder;
  Writer : TWriter;
begin
  BinaryStream:=TMemoryStream.Create;
  try
    BinaryStream.SetSize(0);
    BinaryStream.Position:=0;
    Writer:=TWriter.Create(BinaryStream,4096);
    try
      Writer.Root:=Form;
      c:=0; while c<FSelectedParents.Count do begin
        Writer.WriteSignature;
        { This Freaker stuff gets hairy!  We write a freaker component if we write its holder }
        { but we don't write its holder cos the FORM don't OWN the holder }
        { unless the holder has been selected specifically }
        { so..... we need to scan for freakers by hand }
        Component:=GetEditableComponent(FSelectedParents[c]);
        Writer.WriteComponent(Component);
        inc(c);
      end;
      { now write out freakers whose holders have a PARENT in the list, NOT THOSE whose HOLDERS THEMSELVES }
      { are in the list }
      c:=0; while c<FFreakers.Count do begin
        Component:=FFreakers[c];
        { here is a freaker..... }
        Holder:=(FindHolder(Component) as TComponentHolder);
        If Holder<>nil then begin
          { if holder is child of one of above }
          p:=0; while p<FSelectedParents.Count do begin
            Control:=FSelectedParents[p];
            If (Control<>Holder) and (Control is TWinControl) and
            (Control as TWinControl).ContainsControl(Holder) then begin
              Writer.WriteSignature;
              Writer.WriteComponent(Component);
              break;
            end;
          inc(p);
          end;
        end;
        inc(c);
      end;
    finally
      Writer.Free;
    end;
    BinaryStream.Position:=0;
    { Turn it into text }
    TextStream:=TMemoryStream.Create;
    try
      while BinaryStream.Position<BinaryStream.Size do begin
        ObjectBinaryToText(BinaryStream,TextStream);
      end;
      c:=0;
      { Null terminate the text }
      TextStream.Write(c,1);
      StreamToClipboard(TextStream);
    finally
      TextStream.Free;
    end;
  finally
    BinaryStream.Free;
  end;
end;

{ This does >64K }
Procedure TFormEditor2.StreamToClipboard;
var
  Data: THandle;
  DataPtr: Pointer;
begin
  If OpenClipboard(Application.Handle) then begin
    try
      Data := GlobalAlloc(GMEM_MOVEABLE, Stream.Size+1);
      try
        DataPtr := GlobalLock(Data);
        try
          HMemcpy(DataPtr,Stream.Memory,Stream.Size);
          SetClipboardData(CF_Text, Data);
        finally
          GlobalUnlock(Data);
        end;
      except
        GlobalFree(Data);
        raise;
      end;
    finally
      CloseClipboard;
    end;
  end else Raise Exception.Create('Clipboard is locked by another application');
end;

Procedure TFormEditor2.StreamFromClipboard;
var
  Data: THandle;
  DataPtr: Pointer;
  Size : longint;
begin
  If OpenClipboard(Application.Handle) then begin
    try
      Data := GetClipboardData(CF_TEXT);
      if Data = 0 then Exit;
      DataPtr := GlobalLock(Data);
      try
        Size:=Mystrlen(Data); { can do >64K}
        Stream.SetSize(size);
        HMemcpy(Stream.Memory,DataPtr,size);
      finally
        GlobalUnlock(Data);
      end;
    finally
      CloseClipboard;
    end;
  end else Raise Exception.Create('Clipboard is locked by another application');
end;

Procedure TFormEditor2.ReadComponentsFromStream;
var
  Reader : TReader;
  Name : String;
  Pos : Longint;
begin
  Reader := TReader.Create(Stream, 256);
  try
    Reader.OnSetName:=ForceValidName;
    Reader.Owner:=Owner;
    Reader.parent:=Parent;
    FPastedComponent:=nil;
    FAutoHolder:=True;
    Reader.ReadComponents(Owner,Parent,ReadedForPaste);
  finally
    FAutoHolder:=false;
    Reader.Free;
  end;
end;

Procedure TFormEditor2.ReadedForPaste;
  Function PasteOne(Component : TComponent; PasteControl : TWinControl) : TControl;
  var
    C : Integer;
    Control : Tcontrol;
  begin
    If Component.Owner=nil then
      Form.InsertComponent(Component); { should never happen }
    If not (Component is TControl) then begin
      Control:=FLastHolder;
      { Make Holder paint itself and get itself right! }
      FLastHolder.ComponentMoved;
    end else begin
      Control:=Component as TControl;
      If Control.Parent=nil then
        Control.Parent:=PasteControl;
    end;
    { Cannot do this!  Not till we are loaded! }
    { Cos fixups aren't done yet }
    {    FreakAutoParents(Component,PasteControl);}
    FNewFreakers.Add(Component);
    Result:=Control;
  end;

var
  Control : TControl;
begin
  FPastedComponent:=Component;
  Control:=PasteOne(Component,FPasteControl);
  FixPosition(Control);
  AddSelection(Control);
end;

Procedure TFormEditor2.Paste;
var
  Component : TComponent;
  Control : TControl;
  TextStream, BinaryStream : TMemoryStream;
  Data : THandle;
  DataPtr : Pointer;
  Reader : TReader;
  Writer : TWriter;

begin
  ModifiedForm;
  FPasteControl:=ValidatePasteControl(FPasteControl);
  ClearSelection;
  FNewFreakers.Clear;

  TextStream:=TMemoryStream.Create;

  try
    StreamFromClipboard(TextStream);

    BinaryStream:=TMemoryStream.Create;
    try
      TextStream.Position:=0;

      { this should be one happy copy now }
      While TextStream.Position<TextStream.Size do begin{ cos of null...??SAMFIX? }
        ObjectTextToBinary(TextStream,BinaryStream);
      end;
      { we MUST write an end of list to the stream }
      Writer:=TWriter.Create(BinaryStream,256);
      try
        Writer.WriteListEnd;
      finally
        Writer.Free;
      end;
      BinaryStream.Position:=0;

      { now read the group of components }
      try
        ReadComponentsFromStream(BinaryStream,Form,FPasteControl);
      except
        On E: Exception do
          Application.ShowException(E);
      end;
      FreakNewFreakers(FPasteControl);
      FixZOrder;
    finally
      BinaryStream.Free;
    end;
  finally
    TextStream.Free;
  end;
  Modified; { designer.modified }
end;

Procedure TFormEditor2.Delete;
  procedure KillComponent(Component : TComponent);
  var
    Control : Tcontrol;
  begin
    { This used to be complicted }
    Component.Free;
  end;

var
  c : integer;
  DelList : TList;
begin
  ModifiedForm;
  DelList:=TList.Create;
  Try
    c:=0; while c<FSelectedParents.Count do begin
      DelList.Add(GetEditableComponent(FSelectedParents[c]));
      inc(c);
    end;
    ClearSelection;
    c:=0; while c<DelList.Count do begin
        KillComponent(DelList[c]);
      inc(c);
    end;
  finally
    DelList.Free;
  end;
  Modified;
end;

Procedure TFormEditor2.Cut;
begin
  ModifiedForm;
  Copy;
  Delete;
end;

Procedure TFormEditor2.EmergencyQuit;
begin
  { used to be for infinite message sending loops }
  { not needed now }
{  MainForm.GlobalTickCount:=GetTickCount;}
end;

{ Make sure handles and holders are always on top }
Procedure TFormEditor2.FixZOrder;
var
  ZList : TList;
  Component : TComponent;
  Control : TControl;
  c, Index : Integer;
begin
  c:=0; while c<FTempOwner.ComponentCount do begin
    Component:=FTempOwner.Components[c];
    If (Component is TWinControl) then with Component as TWinControl do
      If Parent=Form then
        BringtoFront;
    If Component is TXControlHandle then with Component as TXControlHandle do SetOnTop;
    inc(c);
  end;
end;

{ This ought not to happen but it might, and it disposes of stuff }
Procedure KillDesigners;
var
  c : integer;
  Designer : TFormEditor2;
  KillList : TList;
begin
  KillList:=TList.Create;
  try
    c:=0; while c<Designers.Count do begin
      Designer:=Designers[c];
      If Designer<>nil then
        KillList.Add(Designer);
      inc(c);
    end;

    c:=0; while c<KillList.Count do begin
      Designer:=KillList[c];
      Designer.Free;
      inc(c);
    end;
  finally
    KillList.Free;
    Designers.Free;
  end;
end;

Procedure TFormEditor2.SetShowHandles;
begin
  FShowHandles:=Value;
  If FVisibleHandles then RefreshAllHandles
end;

Function TFormEditor2.ShowingHandles;
begin
  Result:=FVisibleHandles and FShowhandles;
end;

Procedure TFormEditor2.FormSizePosChanged;
begin
  If FFormLoaded then begin
    ModifiedForm;
    { Now if we are object inspected do a Modified }
    { Maybe do anyway incase it affected lots of things }
    Modified;
  end;
end;

Procedure TFormEditor2.LooseDesignerFocus;
begin
  If not Commit then Inspector.CommitChanges(Self,Commit);
  { otherwise commit is auto }
  Inspector.UseComponents(nil,Self);
end;

Procedure TFormEditor2.GetDesignerFocus;
begin
  If Inspector.ActiveEditor<>Self then
    Editcomponents; { Make the inspector point to us }
end;

Procedure TFormEditor2.SetFilename;
begin
  { Update the form caption if it was the same as the file name }
  If FFilename=Form.Caption then begin
    Form.Caption:=Value;
    FFilename:=Value; { we must set filename before we call modified }
                      { so that knock-on-effect notifies get the real name }
    Modified;
  end else begin
    FFilename:=Value;
    Modified; {SAMFIX not exactly MODIFIED but something to cause people to know the filename has changed }
  end;
end;

Function TFormEditor2.GetFreakInfo;
var
  TryClass : TClass;
  d : String;
begin
  TryClass:=Component.Classtype;
  d:=tryclass.classname;
  Result:=FFreakClassRules.Values[TryClass.Classname];
  While (result='') and (TryClass<>TComponent) do begin
    TryClass:=TryClass.ClassParent;
  d:=tryclass.classname;
    Result:=FFreakClassRules.Values[TryClass.Classname];
  end;
end;

Function TFormEditor2.CanBeFreaked;
begin
  Result:=GetFreakInfo(Component)<>'';
end;

Function TFormEditor2.TryFreak;
var
  Rect : TRect;
  FreakInfo : string;
  FreakProperty : String;
  FreakClasses : String;
  Propinfo : PPropInfo;
  Typedata : PTypedata;
  TrialCandidate : TObject;
  RightClass : Boolean;

begin
  Result:=False;
  RightClass:=False;
  Propinfo:=nil;
  FreakInfo:=GetFreakInfo(Component);
  if FreakInfo='' then exit;

  Split(',',FreakInfo,FreakProperty,FreakClasses);

  If TryInternal then begin
    { Get property info }
    PropInfo:=GetPropinfo(Component.ClassInfo,FreakProperty);
    If PropInfo=nil then exit;
    if PropInfo^.PropType^.Kind <> tkClass then exit;
    TrialCandidate:=TObject(GetOrdProp(Component, PropInfo));
    If TrialCandidate=nil then exit;
    If not (TrialCandidate is TWinControl) then exit;
    Candidate:=TrialCandidate as TWinControl;
    If Candidate=nil then exit;
    { do a quick check and exit, because as we HAVE a candidate which IS }
    { set to the property, it MUST match, so if this is also the visual's parent }
    { then we can exit TRUE and save a LOT of time }
    RightClass:=Candidate=Visual.Parent;
  end;

  If Candidate=nil then Candidate:=Form;

  { update visual }
  { check if candidate is one of supplied classes }
  while (freakClasses<>'') and (RightClass=False) do begin
    Split(',',FreakClasses,FreakInfo,FreakClasses);
    RightClass:=Candidate is FindClass(FreakInfo);
  end;
  If (Candidate=Form) or RightClass then begin
    Rect:=Visual.BoundsRect;
    If MapCoords then begin
      Rect.TopLeft:=Candidate.ScreenToClient(Visual.Parent.ClientToScreen(Rect.TopLeft));
      Rect.BottomRight:=Candidate.ScreenToClient(Visual.Parent.ClientToScreen(Rect.BottomRight));
    end;

    If Visual.Parent<>Candidate then
      Visual.Parent:=Candidate;
    Visual.BoundsRect:=Rect;
    { first check freaking bounds }
    FixPosition(Visual);

    If Propinfo=nil then begin { must have not set it up, do so now }
      PropInfo:=GetPropinfo(Component.ClassInfo,FreakProperty);
      If PropInfo<>nil then
        if PropInfo^.PropType^.Kind <> tkClass then Propinfo:=nil;
    end;

    { now check for class compatabilty with the actual property before we set it }
    If Propinfo<>nil then begin
      If Propinfo^.PropType^.Kind<>tkClass then
        TypeData:=nil
      else
        Typedata:=GetTypeData(PropInfo^.PropType);
      If TypeData=nil then PropInfo:=nil else begin
        If not (Candidate is TypeData^.ClassType) then
          PropInfo:=nil;
      end;
    end;

    If Propinfo<>nil then begin
      If RightClass then
        SetOrdProp(Component, PropInfo, Longint(Candidate))
      else
        SetOrdProp(Component, PropInfo, Longint(nil));

      Result:=True;
      If FFreakers.IndexOf(Component)=-1 then
        FFreakers.Add(Component);
    end;
  end;
end;

{ FreakAuto parents is called when a component is newly inserted }
{ but after fixups are fixed }
{ we try and freak to the freak property first, if that is nil }
{ THEN we freak to the candidate }
Procedure TFormEditor2.FreakAutoParents;
var
  Control : TControl;

begin
  If Component is TControl then
    Control:=Component as TControl
  else
    Control:=FindHolder(Component);
  If Control=nil then exit;

  { we have actual component, and visual representation }
  { Candidate is a candidate for parent }

  If OwnFirst then begin
    If not TryFreak(Component,Control,Candidate,false,map) then
    If not TryFreak(Component,Control,Candidate,true,map) then
    TryFreak(Component,Control,nil,false,map);
  end else begin
    If not TryFreak(Component,Control,Candidate,true,map) then
    If not TryFreak(Component,Control,Candidate,false,map) then
    TryFreak(Component,Control,nil,false,map);
  end;
end;

Procedure TFormEditor2.CheckFreakers;
var
  c : integer;
  Component : TComponent;

  Procedure ValidateFreaking(Component : TComponent);
  var
    Control : TControl;
  begin
    If Component is TControl then
      Control:=Component as TControl
    else begin
      Control:=FindHolder(Component);
      If Control=nil then begin
        Component.Free;
        Exit;
      end;
    end;
    { If freak property is not equal to parent then change parent to freak property }
    { ubless property is nill, then set parent to form }
    If not TryFreak(Component,Control,nil,true,false)
    then TryFreak(Component,Control,nil,false,false);
  end;

begin
  c:=0; while c<FFreakers.Count do begin
    Component:=FFreakers[c];
    If Component<>nil then ValidateFreaking(Component);
    inc(c);
  end;
end;

Procedure TFormEditor2.FreakNewFreakers;
var
  C : Integer;
begin
  c:=0; while C<FNewFreakers.Count do begin
    FreakAutoParents(TComponent(FNewFreakers[c]),PasteControl,false,true);
    inc(c);
  end;
end;

Procedure TFormEditor2.PrepareFreakers;
var
  c : integer;
  Component : TComponent;
  Control : TControl;
begin
  { scan all components and test for freaking }
  { any non-visual will have their holders by now }
  c:=0; while c<Form.ComponentCount do begin
    Component:=Form.Components[c];
    If CanBeFreaked(Component) then begin
      If Component is TControl then
        Control:=Component as TControl
      else begin
        Control:=FindHolder(Component);
      end;
      If not TryFreak(Component,Control,nil,true,false)
      then TryFreak(Component,Control,nil,false,false);
    end;
    inc(c);
  end;
end;

Procedure TFormEditor2.PrepareHolders;
var
  c : Integer;
  component : TComponent;
begin
  c:=0; while c<FtempOwner.ComponentCount do begin
    Component:=FTempOwner.Components[c];
    If Component is TComponentHolder then
      With Component as TComponentHolder do ComponentMoved;
    inc(c);
  end;
end;

Procedure TFormEditor2.CommitChanges;
begin
  { if we are in the inspector then commit any partially enetered changes }
  Inspector.CommitChanges(Self,true);
end;

initialization
  Designers:=TList.Create;
end.


