unit Peedit;

interface

uses Classes, Controls, Winprocs, Wintypes, Sysutils, dsgnintf, stdctrls,
     messages, Forms, graphics, typinfo;

type
  TInputType=(itEdit,itDropDown,itButton,itSubProp,itLabel);

  TCustomVisualPropertyEditor=Class(TWinControl)
  protected
    FChanged : Boolean;
    FSlave : Boolean;
    FOnEditSubProperties : TNotifyEvent;
    FPropertyEditor : TPropertyEditor;
    FPropertyType : TPropertyAttributes;
    FVisualEditor : TControl;
    Procedure PartialEdit(Sender : TObject); virtual;    { called each time item changes }
    Procedure WriteValue; virtual;    { writes edit value back, and makes sure they match if error }
    Function GetEditValue : String; virtual;  { returns current edited value }
    Procedure SetEditValue(Value : String); virtual; { sets current editing value }
    Procedure SetPropertyEditor(Value : TPropertyEditor); virtual;
    Procedure FreePropertyEditor; virtual;
    Procedure WMSize(Var Message : TWMSize); message WM_Size;
    Procedure WMLButtonDblClk(Var Message : TWMLButtonDblClk); message WM_LButtonDblClk;
    Procedure CheckKey(Sender : TObject; var Key : Char); virtual;
    Procedure DoubleClicked; virtual;
    Procedure EditSubProperties(Sender : TObject); virtual;
    Procedure ReleasePropertyEditor; virtual;
    Procedure GetFocus(Sender : TObject); virtual;
    Procedure SetChanged(Value : Boolean); virtual;
    Function GetChanged : Boolean; virtual;
  public
    Constructor Create(AOwner : TComponent); override;
    Destructor Destroy; override;
    Procedure RemoveFocus;
    Procedure CommitChange(Commit : Boolean);
    Procedure Edit; virtual;
    Property PropertyEditor : TPropertyEditor Read FPropertyEditor Write SetPropertyEditor;
    Property EditValue : String Read GetEditValue Write SetEditValue;
    Procedure Renew; virtual;                 { Causes visual value to be updated }
    Procedure SetOnEditSubProperties(Value : TNotifyEvent); virtual;
    Property Slave : Boolean Read FSlave Write FSlave;
    Property Changed : Boolean Read GetChanged Write SetChanged;
    Procedure FinishedEdit(Sender : TObject); virtual;   { called at the end of a change }
  published
    Property OnEditSubProperties : TNotifyEvent Read FOnEditSubProperties Write SetOnEditSubProperties;
  end;

  TVisualPropertyEditor=class(TCustomVisualPropertyEditor)
  protected
    FOnLooseFocus : TNotifyEvent;
    FLabel : TLabel;
    FInputType : TInputType;
    FVisualPropertyEditor : TCustomVisualPropertyEditor;
    Procedure SetPropertyEditor(Value : TPropertyEditor); override;
    Procedure FreePropertyEditor; override;
    Function GetEditValue : String; override;
    Procedure SetEditValue(Value : String); override;
    Procedure SetOnEditSubProperties(Value : TNotifyEvent); override;
    Procedure HideVisualPropertyEditor; virtual;
    Procedure WMSize(Var Message : TWMSize); message WM_Size;
    Procedure RefreshLabel;
    Procedure ObtainFocus(Sender : TObject);
    Procedure GetFocus(Sender : TObject); override;
    Procedure LooseFocus(Sender : TObject); virtual;
    Procedure SetChanged(Value : Boolean); override;
    Function GetChanged : Boolean; override;
  public
    Constructor Create(AOwner : TComponent); override;
    Destructor Destroy; override;
    Property OnLooseFocus : TNotifyEvent Read FOnLooseFocus Write FOnLooseFocus;
    Procedure ShowVisualPropertyEditor; virtual;
  published
    Property TabOrder;
  end;

  TVisualListPropertyEditor=class(TCustomVisualPropertyEditor)
  protected
    FWasKeyPress : Boolean;       { if list change was caused by keypress and not by selecion }
    FStringList : TStringList;    { For sorting lists }
    FDropDown : TComboBox;
    Procedure LoadDropDown(const Value : String);
    Procedure SetPropertyEditor(Value : TPropertyEditor); override;
    Function GetEditValue : String; override;
    Procedure SetEditValue(Value : String); override;
    procedure WndProc(var Message: TMessage); override;
    Procedure DoubleClicked; override;
    Procedure ListChanged(Sender : TObject);     { called each time item changes }
    Procedure CheckKey(Sender : TObject; var Key : Char); override;
    Procedure FixIndex;
  public
    Destructor Destroy; override;
  end;

  TVisualTextPropertyEditor=class(TCustomVisualPropertyEditor)
  protected
    FEditBox : TEdit;
    Procedure SetPropertyEditor(Value : TPropertyEditor); override;
    Function GetEditValue : String; override;
    Procedure SetEditValue(Value : String); override;
  end;

  TVisualEditOnlyPropertyEditor=class(TCustomVisualPropertyEditor)
  protected
    FButton : TButton;
    Procedure SetPropertyEditor(Value : TPropertyEditor); override;
    Function GetEditValue : String; override;
    Procedure SetEditValue(Value : String); override;
    Procedure EditClick(Sender : TObject);
  end;

  TVisualSubPropOnlyPropertyEditor=class(TCustomVisualPropertyEditor)
  protected
    FButton : TButton;
    Procedure SetPropertyEditor(Value : TPropertyEditor); override;
    Function GetEditValue : String; override;
    Procedure SetEditValue(Value : String); override;
    Procedure EditClick(Sender : TObject);
  end;

  TVisualLabelPropertyEditor=class(TCustomVisualPropertyEditor)
  protected
    FLabel : TLabel;
    Procedure SetPropertyEditor(Value : TPropertyEditor); override;
    Function GetEditValue : String; override;
    Procedure SetEditValue(Value : String); override;
    Procedure EditClick(Sender : TObject);
  end;

const
  PWM_OnClick : Word=0;

Function NiceValue(PropertyEditor : TPropertyEditor; Value : String) : String;

Procedure Register;

implementation

Function NiceValue(PropertyEditor : TPropertyEditor; Value : String) : String;
  Function Nicely(S : String) : String;
  var
    FoundUpper : Boolean;
    C : integer;
  begin
    Result:='';
    FoundUpper:=False;
    c:=1; while c<length(s) do begin
      if not((s[c]>='a') and (s[c]<='z')) then begin
        Result:=Copy(s,c,length(s)-c+1);
        FoundUpper:=True;
        Break;
      end;
      inc(c);
    end;
    If not FoundUpper then Result:=S;
  end;

var
  TI : PTypeInfo;
begin
  { for enumerated types make it look nice }
  { also check by hand for registered integer types, TColor, TCursor, TShortCut }
  { we can't get a list of registered integers so we check by hand for known types }
  If PropertyEditor is TColorProperty then begin
    Result:=Nicely(Value);
  end else begin
    TI:=PropertyEditor.GetProptype;
    Case TI^.Kind of
      tkEnumeration: begin
        Result:=Nicely(Value);
      end;
    else
      Result:=Value;
    end;
  end;
end;

Function GetMessageId(const s : String) : Word;
var
  Buffer : Array[0..255] of char;
begin
  Result:=RegisterWindowMessage(StrPCopy(Buffer,s));
end;

Procedure TCustomVisualPropertyEditor.FinishedEdit;
begin
  If not (csDesigning in ComponentState) then begin
    If Changed and (FPropertyEditor<>nil) then
      WriteValue;
  end;
end;

Procedure TCustomVisualPropertyEditor.Edit;
begin
  if not (csDesigning in ComponentState) then begin
    If FPropertyEditor<>nil then begin
      FPropertyEditor.Edit;
      FPropertyEditor.Designer.Modified; {SAMFIX?? is this done by edit?}
    end;
  end;
end;

Procedure TCustomVisualPropertyEditor.SetChanged;
begin
  FChanged:=Value;
end;

Function TCustomVisualPropertyEditor.GetChanged;
begin
  Result:=FChanged;
end;

Procedure TCustomVisualPropertyEditor.CommitChange;
begin
  If Commit then begin
    FinishedEdit(Self);
  end else begin
    FChanged:=False;
    Renew;
    RemoveFocus;
  end;
end;

Procedure TCustomVisualPropertyEditor.RemoveFocus;
var
  Form: TForm;
begin
  Form := GetParentForm(Self);
  if Form <> nil then Form.DefocusControl(Self, False);
end;

Procedure TCustomVisualPropertyEditor.EditSubProperties;
begin
  if not (csDesigning in ComponentState) then begin
    If paSubProperties in FPropertyType then begin
      If Assigned(FOnEditSubProperties) then
        FOnEditSubProperties(FPropertyEditor);
    end;
  end;
end;

Procedure TCustomVisualPropertyEditor.PartialEdit;
begin
  if not (csDesigning in ComponentState) then begin
    If (FPropertyEditor<>nil) then
      if (paAutoUpdate in FPropertyType) then
        WriteValue;
  end;
end;

Procedure TCustomVisualPropertyEditor.WriteValue;
begin
  if not (csDesigning in ComponentState) then begin
    try
      begin
        FPropertyEditor.Value:=EditValue;
        FPropertyEditor.Designer.Modified;
      end;
    finally
      Changed:=False;
      If EditValue<>FPropertyEditor.Value then
        EditValue:=FPropertyEditor.Value;
    end;
  end;
end;

Procedure TCustomVisualPropertyEditor.Renew;
begin
  if not (csDesigning in ComponentState) then begin
    If FPropertyEditor<>nil then
      If EditValue<>FPropertyEditor.Value then
        EditValue:=FPropertyEditor.Value;
      Changed:=False;
  end;
end;

Procedure TCustomVisualPropertyEditor.SetEditValue;
begin
end;

Function TCustomVisualPropertyEditor.GetEditValue;
begin
end;

Procedure TCustomVisualPropertyEditor.WMSize;
begin
  Inherited;
  If FVisualEditor<>nil then
    FVisualEditor.SetBounds(0,0,Width,Height);
end;

Procedure TCustomVisualPropertyEditor.WMLButtonDblClk;
begin
  if csDesigning in ComponentState then
    inherited
  else
    if FPropertyEditor<>nil then
      DoubleClicked
    else
      Inherited;
end;

Procedure TCustomVisualPropertyEditor.DoubleClicked;
begin
  if not (csDesigning in ComponentState) then begin
    If FPropertyEditor<>nil then
      Edit;
  end;
end;

Constructor TVisualPropertyEditor.Create;
begin
  Inherited Create(AOwner);
  If csDesigning in ComponentState then begin
    Width:=121;
    Height:=24;
    FVisualEditor:=TEdit.Create(Self);
    FVisualEditor.SetBounds(0,0,Width,Height);
    InsertControl(FVisualEditor);
  end else begin
    FLabel:=TLabel.Create(Self);
    FLabel.SetBounds(0,0,Width,Height);
    InsertControl(FLabel);
    FLabel.AutoSize:=False;
    FLabel.FocusControl:=Self;
    FLabel.OnClick:=ObtainFocus;
    FLabel.Color:=clWhite;
{    OnEnter:=Self.GetFocus;}
    OnExit:=Self.LooseFocus;
  end;
  TabStop:=False;{x}
end;

Destructor TVisualPropertyEditor.Destroy;
begin
  FLabel.Free;
  Inherited Destroy;
end;

Procedure TVisualPropertyEditor.ObtainFocus;
begin
  SetFocus;
end;

Procedure TCustomVisualPropertyEditor.GetFocus;
begin
  If FVisualEditor<>nil then
    If FVisualEditor is TWinControl then
      (FVisualEditor as TWinControl).SetFocus;
end;

Procedure TVisualPropertyEditor.GetFocus;
begin
  ShowVisualPropertyEditor;
  If FVisualPropertyEditor<>nil then
    FVisualPropertyEditor.SetFocus;
end;

Procedure TVisualPropertyEditor.LooseFocus;
begin
  FinishedEdit(Self);
{  GetParentForm(Self).DefocusControl(FVisualPropertyEditor,True);
  HideVisualPropertyEditor;
  If assigned(FOnLooseFocus) then
    FOnLooseFocus(Self);}
end;

Procedure TVisualPropertyEditor.SetChanged;
begin
  Inherited SetChanged(Value);
  If FVisualPropertyeditor<>nil then
    FVisualPropertyEditor.Changed:=Value;
end;

Function TVisualPropertyEditor.GetChanged;
begin
  If FVisualPropertyEditor<>nil then
    FChanged:=FVisualPropertyEditor.Changed;
  Result:=FChanged;
end;

Constructor TCustomVisualPropertyEditor.Create;
begin
  Inherited Create(AOwner);
  UpdateBoundsRect(Rect(0,0,121,24));
  TabStop:=False;
{  OnEnter:=Self.GetFocus;}
end;

Destructor TCustomVisualPropertyEditor.Destroy;
begin
  If not Slave then
    FreePropertyEditor;
  Inherited Destroy;
end;

Procedure TCustomVisualPropertyEditor.CheckKey;
begin
  if not (csDesigning in ComponentState) then begin
    If paReadOnly in FPropertyType then begin
      If key=#13 then
        FPropertyEditor.Edit
      else
        MessageBeep(Word(-1));
      Key:=#0;
    end else  begin
      Changed:=true;
      case key of
        #13: begin
          FinishedEdit(Sender);
          Key:=#0;
        end;
      end;
    end;
  end;
end;

Procedure TCustomVisualPropertyEditor.FreePropertyEditor;
begin
  if not slave then
    FPropertyEditor.Free;
  FPropertyEditor:=nil;
end;

Procedure TCustomVisualPropertyEditor.ReleasePropertyEditor;
begin
  FPropertyEditor:=nil;
end;

Procedure TCustomVisualPropertyEditor.SetPropertyEditor;
begin
  If not slave then
    FreePropertyEditor;
  FPropertyEditor:=Value;
  If FPropertyEditor<>nil then
    FPropertyType:=FPropertyEditor.GetAttributes;
  Changed:=False;
end;

Procedure TCustomVisualPropertyeditor.SetOnEditSubProperties;
begin
  FOnEditSubProperties:=Value;
end;

Procedure TVisualPropertyeditor.SetOnEditSubProperties;
begin
  If FVisualPropertyEditor<>nil then
    FVisualPropertyEditor.OnEditSubProperties:=Value;
  FOnEditSubProperties:=Value;
end;

Function TVisualPropertyEditor.GetEditValue;
begin
  if not (csDesigning in ComponentState) then begin
    If (FVisualPropertyEditor<>nil) then
      Result:=FVisualPropertyEditor.EditValue
    else
      Result:='Error';
  end else
    Result:='DesignMode';
end;

Procedure TVisualPropertyEditor.WMSize;
begin
  Inherited;
  If FLabel<>nil then
    FLabel.SetBounds(0,0,Width,Height);
end;

Procedure TVisualPropertyEditor.SetEditValue;
begin
  if not (csDesigning in ComponentState) then begin
    If (FVisualPropertyEditor<>nil) then
      FVisualPropertyEditor.EditValue:=Value;
    FLabel.Caption:=Value;
  end;
end;

Procedure TVisualPropertyEditor.FreePropertyEditor;
begin
  If not slave then
    FPropertyEditor.Free;
  FPropertyEditor:=nil;
  FVisualEditor.Free;
  FVisualEditor:=nil;
end;

Procedure TVisualPropertyEditor.SetpropertyEditor;
begin
  Inherited SetpropertyEditor(Value);
  RefreshLabel;
end;

Procedure TVisualPropertyEditor.RefreshLabel;
begin
  If (FLabel<>nil) and (FPropertyEditor<>nil) then begin
    FLabel.Caption:=FPropertyEditor.Value;
    Changed:=False;
  end;
end;

Procedure TVisualPropertyEditor.HideVisualPropertyEditor;
begin
  RefreshLabel;
  If (FVisualEditor<>nil) then
    If FVisualEditor is TCustomVisualPropertyEditor then
      (FVisualEditor as TCustomVisualPropertyEditor).ReleasePropertyEditor;
  FVisualEditor.Free;
  FVisualEditor:=nil;
end;

Procedure TVisualPropertyEditor.ShowVisualPropertyEditor;
begin
  If FVisualPropertyEditor<>nil then
      exit;
  if not (csDesigning in ComponentState) and (FPropertyEditor<>nil) then begin
    If (paValueList in FPropertyType) or (paSortList in FPropertyType) then
      { Must be a list }
      FInputType:=itDropDown
    else If (paDialog in FPropertyType) then
      FInputType:=itButton
    else If (paSubProperties in FPropertyType) then
      FInputType:=itSubProp
    else if not (paReadOnly in FPropertyType) then
      FInputType:=itEdit
    else begin
      FInputType:=itLabel;
    end;
    case FInputType of
      itButton: begin
        FVisualPropertyEditor:=TVisualEditOnlyPropertyEditor.Create(Self);
        FVisualPropertyEditor.Slave:=True;
        FVisualPropertyEditor.SetBounds(0,0,Width,Height);
        InsertControl(FVisualPropertyEditor);
        FVisualEditor:=FVisualPropertyEditor;
        FVisualPropertyEditor.PropertyEditor:=FPropertyEditor;
        FVisualPropertyEditor.OnEditSubProperties:=OnEditSubProperties;
      end;
      itSubProp: begin
        FVisualPropertyEditor:=TVisualSubPropOnlyPropertyEditor.Create(Self);
        FVisualPropertyEditor.Slave:=True;
        FVisualPropertyEditor.SetBounds(0,0,Width,Height);
        InsertControl(FVisualPropertyEditor);
        FVisualEditor:=FVisualPropertyEditor;
        FVisualPropertyEditor.PropertyEditor:=FPropertyEditor;
        FVisualPropertyEditor.OnEditSubProperties:=OnEditSubProperties;
      end;
      itDropDown: begin
        FVisualPropertyEditor:=TVisualListPropertyEditor.Create(Self);
        FVisualPropertyEditor.Slave:=True;
        FVisualPropertyEditor.SetBounds(0,0,Width,Height);
        InsertControl(FVisualPropertyEditor);
        FVisualEditor:=FVisualPropertyEditor;
        FVisualPropertyEditor.PropertyEditor:=FPropertyEditor;
        FVisualPropertyEditor.OnEditSubProperties:=OnEditSubProperties;
      end;
      itEdit: begin
        FVisualPropertyEditor:=TVisualTextPropertyEditor.Create(Self);
        FVisualPropertyEditor.Slave:=True;
        FVisualPropertyEditor.SetBounds(0,0,Width,Height);
        InsertControl(FVisualPropertyEditor);
        FVisualEditor:=FVisualPropertyEditor;
        FVisualPropertyEditor.PropertyEditor:=FPropertyEditor;
        FVisualPropertyEditor.OnEditSubProperties:=OnEditSubProperties;
      end;
      itLabel: begin
        FVisualPropertyEditor:=TVisualLabelPropertyEditor.Create(Self);
        FVisualPropertyEditor.Slave:=True;
        FVisualPropertyEditor.SetBounds(0,0,Width,Height);
        InsertControl(FVisualPropertyEditor);
        FVisualEditor:=FVisualPropertyEditor;
        FVisualPropertyEditor.PropertyEditor:=FPropertyEditor;
        FVisualPropertyEditor.OnEditSubProperties:=OnEditSubProperties;
      end;
    end;
    Changed:=False;
  end;
end;

Procedure TVisualListPropertyEditor.LoadDropDown;
begin
  FStringList.AddObject(NiceValue(FPropertyEditor,Value),TObject(NewStr(Value)));
end;

Procedure TVisualListPropertyEditor.WndProc;
var
  P : PString;
begin
  If Message.Msg=PWM_OnClick then begin
    P:=Pointer(Message.LParam);
    If FDropDown<>nil then
      FDropDown.Text:=P^;
    DisposeStr(P);
  end else
    Inherited WndProc(Message);
end;

{ this function is given a canonical value and turns to a nice value }
Procedure TVisualListPropertyEditor.SetEditValue;
var
  P : PString;
begin
  if not (csDesigning in ComponentState) then begin
    If FDropDown<>nil then
      Value:=NiceValue(FPropertyEditor,FPropertyEditor.Value);
      FDropDown.Text:=Value;
    { There is a weird bug in TComboBox whereby if you change }
    { .Text in the OnClick event, its DISPLAY gets changed back! }
    { So here we will send a message to ourselves to fix it }
    P:=NewStr(Value);
    PostMessage(Handle,PWM_OnClick,0,longint(P));
  end;
end;

{ This function should return the canonical value, so we need to reverse translate }
Function TVisualListPropertyEditor.getEditvalue;
begin
  if not (csDesigning in ComponentState) then begin
    If FDropDown<>nil then begin{ translate the value }
      FixIndex;
      If (FDropDown.ItemIndex>=FDropDown.Items.Count) or (FDropDown.ItemIndex<0) then
        Result:=FDropDown.Text
      else
        try
          Result:=PString(FStringList.Objects[FDropDown.ItemIndex])^;
        except
          Result:=FDropDown.Text;
        end;
    end else
      Result:='Not initialised';
  end else
    Result:='Design';
end;

Procedure TVisualListPropertyEditor.FixIndex;
var
  NewIndex : Integer;
  c : integer;
begin
  If (FDropDown.ItemIndex>=FDropDown.Items.Count) or (FDropDown.ItemIndex<0) then begin
    NewIndex:=FDropDown.Items.IndexOf(FDropDown.Text);
    If NewIndex=-1 then begin
      { have to do a comparetext search }
      c:=0; while c<FDropDown.Items.Count do begin
        If CompareText(FDropDown.Text,FDropDown.Items[c])=0 then begin
          NewIndex:=c;
          break;
        end;
        inc(c);
      end;
    end;
    If NewIndex<>-1 then FDropDown.ItemIndex:=NewIndex;
  end;
end;

Procedure TVisualListPropertyEditor.ListChanged;
begin
  Changed:=true;
  If not FWasKeyPress then
    WriteValue;
  FWasKeyPress:=False;
end;

Procedure TVisualListPropertyEditor.SetPropertyEditor;
begin
  Inherited SetPropertyEditor(Value);
  If FPropertyEditor<>nil then begin
    If (paValueList in FPropertyType) or (paSortList in FPropertyType) then begin
      FDropDown:=TComboBox.Create(Self);
      FDropDown.SetBounds(0,0,Width,height);
      FVisualEditor:=FDropDown;
      InsertControl(FDropDown);
      { Now add items to dropdown }
      FStringList:=TStringList.Create;
      If paSortList in FPropertyType then
        FStringList.Sorted:=True;
      FStringList.BeginUpdate;
      FPropertyEditor.GetValues(LoadDropDown);
      FStringList.EndUpdate;
      FDropDown.Items.AddStrings(FStringList);
      Renew; { Load fresh vaue }
      FDropDown.OnChange:=Self.ListChanged;
      FDropdown.OnKeyPress:=Self.CheckKey;
      FDropDown.OnClick:=Self.FinishedEdit;
    end else if not slave then
      FreePropertyEditor;
  end;
end;

Procedure TVisualListPropertyEditor.CheckKey;
begin
  FWasKeyPress:=true;
  Inherited CheckKey(Sender,Key);
end;

Procedure TVisualListPropertyEditor.DoubleClicked;
var
  c : Integer;
begin
  If (FDropDown=nil) or (csDesigning in ComponentState) then exit;
  c:=FDropDown.Items.IndexOf(FDropDown.Text);
  inc(c);
  if c>=FDropDown.Items.Count then c:=0;
  FDropDown.Text:=FDropDown.Items[c];
  Changed:=true;
end;

Destructor TVisualListpropertyEditor.Destroy;
var
  C : integer;
begin
  { Free all those newstr's in the dropdown box }
  try
    c:=0; while C<FStringList.Count do begin
      DisposeStr(PString(FStringList.Objects[c]));
      inc(c);
    end;
  finally
    FStringList.Free;
    Inherited Destroy;
  end;
end;

Procedure TVisualTextPropertyEditor.SetEditValue;
begin
  If FEditBox<>nil then
    FEditBox.Text:=Value;
end;

Function TVisualTextPropertyEditor.getEditvalue;
begin
  If FEditBox<>nil then
    Result:=FEditBox.Text
  else
    Result:='Not initialised';
end;

Procedure TVisualTextPropertyEditor.SetPropertyEditor;
begin
  Inherited SetPropertyEditor(Value);
  If FPropertyEditor<>nil then begin
    FEditBox:=TEdit.Create(Self);
    FEditBox.SetBounds(0,0,Width,Height);
    FVisualEditor:=FEditBox;
    InsertControl(FEditBox);
    Renew;
    FEditBox.OnChange:=Self.PartialEdit;
    FEditBox.OnKeyPress:=Self.CheckKey;
  end;
end;

Procedure TVisualEditOnlyPropertyEditor.SetEditValue;
begin
  If FButton<>nil then
    FButton.Caption:='Edit '+FPropertyEditor.GetName;
end;

Function TVisualEditOnlyPropertyEditor.getEditvalue;
begin
  If FButton<>nil then
    Result:=FButton.Caption
  else
    Result:='Not initialised';
end;

Procedure TVisualEditOnlyPropertyEditor.SetPropertyEditor;
begin
  Inherited SetPropertyEditor(Value);
  If FPropertyEditor<>nil then begin
    FButton:=TButton.Create(Self);
    FButton.SetBounds(0,0,Width,Height);
    FVisualEditor:=FButton;
    InsertControl(FButton);
    Renew;
    FButton.OnClick:=Self.EditClick;
  end;
end;

Procedure TVisualEditOnlyPropertyEditor.EditClick;
begin
  Edit;
end;

Procedure TVisualSubPropOnlyPropertyEditor.SetEditValue;
begin
  If FButton<>nil then
    FButton.Caption:=Value;
end;

Function TVisualSubPropOnlyPropertyEditor.getEditvalue;
begin
  If FButton<>nil then
    Result:=FButton.Caption
  else
    Result:='Not initialised';
end;

Procedure TVisualSubPropOnlyPropertyEditor.SetPropertyEditor;
begin
  Inherited SetPropertyEditor(Value);
  If FPropertyEditor<>nil then begin
    FButton:=TButton.Create(Self);
    FButton.SetBounds(0,0,Width,Height);
    FVisualEditor:=FButton;
    InsertControl(FButton);
    Renew;
    FButton.OnClick:=Self.EditSubProperties;
  end;
end;

Procedure TVisualSubPropOnlyPropertyEditor.EditClick;
begin
  Edit;
end;

Procedure TVisualLabelPropertyEditor.SetEditValue;
begin
  If FLabel<>nil then
    FLabel.Caption:=Value;
end;

Function TVisualLabelPropertyEditor.getEditvalue;
begin
  If FLabel<>nil then
    Result:=FLabel.Caption
  else
    Result:='Not initialised';
end;

Procedure TVisualLabelPropertyEditor.SetPropertyEditor;
begin
  Inherited SetPropertyEditor(Value);
  If FPropertyEditor<>nil then begin
    FLabel:=TLabel.Create(Self);
    FLabel.SetBounds(0,0,Width,Height);
    FVisualEditor:=FLabel;
    InsertControl(FLabel);
    Renew;
  end;
end;

Procedure TVisualLabelPropertyEditor.EditClick;
begin
  Edit;
end;

Procedure Register;
begin
  RegisterComponents('CSL Extras',[TVisualPropertyEditor]);
end;

begin
  PWM_OnClick:=GetMessageId(Application.ExeName+':OnClickBugFix');
end.


