{$F+}
unit Regfake;
{ This unit supports RegisterComponents to add your own components to the palette }

interface

uses  classes, sysutils, winprocs, wintypes, forms, fedutils;

type
  TRegistrationPage=class(TObject)
  Protected
    FName : PString;
    FComponents : TList;
    Function GetName : String;
    Procedure Setname(Name : String);
    Function getComponentCount : Integer;
  public
    Constructor Create;
    Destructor Destroy;
    Property Name : String Read GetName Write Setname;
    Property ComponentCount : Integer read getComponentCount;
    Function GetComponent(Value : Integer) : TComponentclass;
    Procedure Addcomponent(Component : TComponentclass);
    Function GetBitmap(Value : Integer) : HBitmap;
  end;

  TRegistrationList=class(TObject)
  Protected
    FPages : TList;
    Function GetPageCount : Integer;
  Public
    Constructor Create;
    Destructor Destroy;
    Property PageCount : Integer Read GetPageCount;
    Function FindPage(PageName : String) : TRegistrationPage;
    Function GetPage(Index : Integer) : TRegistrationPage;
    Procedure Register(PageName : String; Component : TComponentClass);
    Function AddPage(PageName : String) : TRegistrationPage;
  end;

var
  RegistrationPages : TRegistrationList;

implementation

Constructor TRegistrationPage.Create;
begin
  Inherited Create;
  FComponents:=TList.Create;
end;

Destructor TRegistrationPage.Destroy;
begin
  FComponents.Free;
  Inherited Destroy;
end;

Function TRegistrationPage.getComponentCount;
begin
  Result:=FComponents.Count;
end;

Function TRegistrationPage.getComponent;
begin
  Result:=FComponents[value];
end;

Function TRegistrationPage.getBitMap;
begin
  Result:=GetbestBitmap(GetComponent(Value));
end;

Function TRegistrationPage.GetName;
begin
  if FName=nil then
    Result:=''
  else
    Result:=Fname^;
end;

Procedure TRegistrationPage.Setname;
begin
  If FName<>nil then
    DisposeStr(Fname);
  FName:=NewStr(Name);
end;

Procedure TRegistrationpage.AddComponent;
begin
  FComponents.Add(Component);
end;

Constructor TRegistrationList.Create;
begin
  Inherited create;
  FPages:=TList.Create;
end;

Destructor TRegistrationList.Destroy;
var
  c : Integer;
begin
  C:=0; while c<FPages.Count do begin
    TRegistrationPage(FPages[c]).Free;
    inc(c);
  end;
  FPages.Free;
  Inherited Destroy;
end;

Function TRegistrationList.GetPage;
begin
  Result:=TRegistrationPage(FPages[Index]);
end;

Function TRegistrationList.FindPage;
var
  c : Integer;
  Page : TRegistrationPage;
begin
  Result:=nil;
  c:=0; while c<FPages.Count do begin
    Page:=TRegistrationPage(FPages[c]);
    If Page.Name=Pagename then begin
      Result:=Page;
      break;
    end;
    inc(c);
  end;
end;

Procedure TRegistrationList.Register;
var
  Page : TRegistrationPage;
begin
  Page:=FindPage(PageName);
  If Page=nil then
    Page:=AddPage(PageName);
  Page.AddComponent(Component);
end;


Function TRegistrationList.AddPage;
begin
  Result:=TRegistrationPage.Create;
  If Result<>nil then begin
    Result.Name:=PageName;
    FPages.Add(Result);
  end;
end;

Function TRegistrationList.GetPageCount;
begin
  Result:=FPages.Count;
end;

Procedure RegisterComponents(const Page: string; ComponentClasses: array of TComponentClass);
var
  c : Integer;
begin
  c:=Low(ComponentClasses);
  while c<=High(ComponentClasses) do begin
    RegistrationPages.Register(Page,ComponentClasses[c]);
    classes.RegisterClass(ComponentClasses[c]);
    inc(c);
  end;
end;

Procedure RegisterNoIcon(ComponentClasses: array of TComponentClass);
var
  c : Integer;
begin
  c:=Low(ComponentClasses);
  while c<=High(ComponentClasses) do begin
    classes.RegisterClass(ComponentClasses[c]);
    inc(c);
  end;
end;

Procedure FreeRegistrationList;
begin
  RegistrationPages.Free;
end;

begin
  RegistrationPages:=TRegistrationList.Create;
  AddExitProc(FreeRegistrationList);
  @RegisterComponentsProc:=@RegisterComponents;
  @RegisterNoIconProc:=@RegisterNoIcon;
end.
